import numpy as np
from loguru import logger

from openamundsen import constants as c
from openamundsen import meteo

from .soiltexture import SoilTextureClass

# Default soil water characteristics for different soil types (from Table 19 in Allen et al. (1998))
DEFAULT_SOIL_WATER_CONTENTS_AT_FIELD_CAPACITY = {  # m3 m-3
    SoilTextureClass.SAND: (0.07 + 0.17) / 2,
    SoilTextureClass.LOAMY_SAND: (0.11 + 0.19) / 2,
    SoilTextureClass.SANDY_LOAM: (0.18 + 0.28) / 2,
    SoilTextureClass.LOAM: (0.20 + 0.30) / 2,
    SoilTextureClass.SILT_LOAM: (0.22 + 0.36) / 2,
    SoilTextureClass.SILT: (0.28 + 0.36) / 2,
    SoilTextureClass.SILT_CLAY_LOAM: (0.30 + 0.37) / 2,
    SoilTextureClass.SILTY_CLAY: (0.30 + 0.42) / 2,
    SoilTextureClass.CLAY: (0.32 + 0.40) / 2,
}
DEFAULT_SOIL_WATER_CONTENTS_AT_WILTING_POINT = {  # m3 m-3
    SoilTextureClass.SAND: (0.02 + 0.07) / 2,
    SoilTextureClass.LOAMY_SAND: (0.03 + 0.10) / 2,
    SoilTextureClass.SANDY_LOAM: (0.06 + 0.16) / 2,
    SoilTextureClass.LOAM: (0.07 + 0.17) / 2,
    SoilTextureClass.SILT_LOAM: (0.09 + 0.21) / 2,
    SoilTextureClass.SILT: (0.12 + 0.22) / 2,
    SoilTextureClass.SILT_CLAY_LOAM: (0.17 + 0.24) / 2,
    SoilTextureClass.SILTY_CLAY: (0.17 + 0.29) / 2,
    SoilTextureClass.CLAY: (0.20 + 0.24) / 2,
}
DEFAULT_READILY_EVAPORABLE_WATER = {  # kg m-2
    SoilTextureClass.SAND: (2.0 + 7.0) / 2,
    SoilTextureClass.LOAMY_SAND: (4.0 + 8.0) / 2,
    SoilTextureClass.SANDY_LOAM: (6.0 + 10.0) / 2,
    SoilTextureClass.LOAM: (8.0 + 10.0) / 2,
    SoilTextureClass.SILT_LOAM: (8.0 + 11.0) / 2,
    SoilTextureClass.SILT: (8.0 + 11.0) / 2,
    SoilTextureClass.SILT_CLAY_LOAM: (8.0 + 11.0) / 2,
    SoilTextureClass.SILTY_CLAY: (8.0 + 12.0) / 2,
    SoilTextureClass.CLAY: (8.0 + 12.0) / 2,
}

# Reduce depletion fractions for fine textured soils by 5-10% and increase by 5-10% for coarse
# textured soils ([1], p. 167)
DEFAULT_DEPLETION_FRACTION_ADJUSTMENTS = {
    SoilTextureClass.SAND: 1.10,
    SoilTextureClass.LOAMY_SAND: 1.075,
    SoilTextureClass.SANDY_LOAM: 1.05,
    SoilTextureClass.LOAM: 1.025,
    SoilTextureClass.SILT_LOAM: 1.0,
    SoilTextureClass.SILT: 0.975,
    SoilTextureClass.SILT_CLAY_LOAM: 0.95,
    SoilTextureClass.SILTY_CLAY: 0.925,
    SoilTextureClass.CLAY: 0.90,
}


class EvapotranspirationModel:
    """
    Evapotranspiration model following [1].

    References
    ----------
    .. [1] Allen, R.G., Pereira, L.S., Raes, D., et al. (1998). Crop
       Evapotranspiration-Guidelines for Computing Crop Water Requirements-FAO
       Irrigation and Drainage Paper 56. FAO, Rome, 300(9): D05109.
       http://www.fao.org/3/x0490e/x0490e00.htm
       https://www.researchgate.net/publication/284300773_FAO_Irrigation_and_drainage_paper_No_56
    """

    def __init__(self, model):
        self.model = model
        s = model.state

        s_et = s.add_category("evapotranspiration")
        s_et.add_variable("soil_texture", long_name="Soil texture class", dtype=int, retain=True)  # TODO: move to base or soil group eventually # fmt: skip # noqa: E501
        s_et.add_variable("evaporation", "kg m-2", "Evaporation")
        s_et.add_variable("transpiration", "kg m-2", "Transpiration")
        s_et.add_variable("evapotranspiration", "kg m-2", "Evapotranspiration")
        s_et.add_variable("et_ref", "kg m-2", "Reference evapotranspiration")
        s_et.add_variable("ref_albedo", "1", "Reference surface albedo", retain=True)
        s_et.add_variable("ref_emissivity", "1", "Reference emissivity", retain=True)
        s_et.add_variable("ref_net_radiation", "W m-2", "Reference net radiation")
        s_et.add_variable(
            "soil_heat_flux",
            "W m-2",
            "Soil heat flux beneath the grass reference surface",
        )
        s_et.add_variable("crop_coeff", "1", "Crop coefficient")
        s_et.add_variable("basal_crop_coeff", "1", "Basal crop coefficient")
        s_et.add_variable("evaporation_coeff", "1", "Evaporation coefficient")
        s_et.add_variable("water_stress_coeff", "1", "Water stress coefficient")
        s_et.add_variable("clim_corr", "1", "Climate correction term", retain=True)
        s_et.add_variable(
            "cum_soil_surface_depletion",
            "kg m-2",
            "Cumulative evaporation from the soil surface layer",
            retain=True,
        )
        s_et.add_variable(
            "cum_root_zone_depletion",
            "kg m-2",
            "Cumulative evapotranspiration from the root zone",
            retain=True,
        )
        s_et.add_variable("total_evaporable_water", "kg m-2", "Total evaporable water", retain=True)
        s_et.add_variable("total_available_water", "kg m-2", "Total available water", retain=True)
        s_et.add_variable(
            "readily_evaporable_water",
            "kg m-2",
            "Readily evaporable water",
            retain=True,
        )
        s_et.add_variable(
            "readily_available_water",
            "kg m-2",
            "Readily available water",
            retain=True,
        )
        s_et.add_variable("deep_percolation", "kg m-2", "Deep percolation")
        s_et.add_variable(
            "deep_percolation_evaporation_layer",
            "kg m-2",
            "Deep percolation from the evaporation layer",
            retain=True,
        )
        s_et.add_variable(
            "sealed_interception",
            "kg m-2",
            "Interception for sealed surfaces",
            retain=True,
        )

    def initialize(self):
        model = self.model
        roi = model.grid.roi
        s = self.model.state
        s_et = s.evapotranspiration

        # Prepare unique land cover classes occurring in the model domain and their associated pixel
        # locations
        lccs = np.unique(s.land_cover.land_cover[roi])
        lccs = lccs[lccs > 0]
        lccs = set(lccs) & set(model.config.land_cover.classes.keys())  # calculate ET only for land cover classes with set parameters # fmt: skip # noqa: E501
        self.land_cover_class_pixels = {}
        for lcc in lccs:
            self.land_cover_class_pixels[lcc] = (s.land_cover.land_cover == lcc) & roi

        self._climate_correction()

        # Initialize depletion fractions
        depletion_fraction = np.full((model.grid["rows"], model.grid["cols"]), np.nan)
        for lcc, pos in self.land_cover_class_pixels.items():
            depletion_fraction[pos] = model.config["land_cover"]["classes"][lcc][
                "depletion_fraction"
            ]

        # Calculate total evaporable water (eq. (73)), initialize readily evaporable water, and
        # begin calculation of total available water (eq. (82)) and initial root zone depletion (eq.
        # (87))
        stcs = np.unique(s_et.soil_texture[roi])
        stcs = stcs[stcs > 0]
        for stc in stcs:
            pos = s_et.soil_texture == stc
            swc_field_cap = DEFAULT_SOIL_WATER_CONTENTS_AT_FIELD_CAPACITY[stc]
            swc_wilting_point = DEFAULT_SOIL_WATER_CONTENTS_AT_WILTING_POINT[stc]

            s_et.total_evaporable_water[pos] = (
                1000
                * (swc_field_cap - 0.5 * swc_wilting_point)
                * model.config.evapotranspiration.surface_soil_layer_evaporation_depth
            )
            s_et.readily_evaporable_water[pos] = DEFAULT_READILY_EVAPORABLE_WATER[stc]

            # In the calculation of TAW the multiplication by the rooting depth is missing here (as
            # this is a land cover specific parameter and not soil specific) - follows below
            s_et.total_available_water[pos] = (
                1000 * (swc_field_cap - swc_wilting_point)
                # * rooting_depth
            )

            # Same for the calculation of initial root zone depletion
            swc = swc_field_cap  # assume root zone is near field capacity following heavy rain or irrigation # noqa: E501
            s_et.cum_root_zone_depletion[pos] = (  # eq. (87)
                1000 * (swc_field_cap - swc)
                # * rooting_depth
            )

            # Adjust depletion fraction depending on soil type ([1], p. 167)
            depletion_fraction[pos] *= DEFAULT_DEPLETION_FRACTION_ADJUSTMENTS[stc]

        # Finish calculation of TAW (eq. (82)) and root zone depletion (eq. (87)) and calculate
        # readily available water (eq. (83))
        # (depletion fractions are assumed constant; adjustment using ETc as suggested in [1] (p.
        # 162) is not performed here)
        for lcc, pos in self.land_cover_class_pixels.items():
            rooting_depth = model.config["land_cover"]["classes"][lcc]["rooting_depth"]
            s_et.total_available_water[pos] *= rooting_depth
            s_et.cum_root_zone_depletion[pos] *= rooting_depth
            s_et.readily_available_water[pos] = (
                depletion_fraction[pos] * s_et.total_available_water[pos]
            )

        # Set D_e to 0 at the start of the model run, i.e., assume the topsoil is near field
        # capacity following a heavy rain or irrigation
        s_et.cum_soil_surface_depletion[roi] = 0.0

        s_et.deep_percolation_evaporation_layer[roi] = 0.0

        # Initialize reference albedo, reference emissivity and sealed interception
        for lcc, pos in self.land_cover_class_pixels.items():
            if model.config["land_cover"]["classes"][lcc].get("is_sealed", False):
                s_et.ref_albedo[pos] = model.config.evapotranspiration.sealed_albedo
                s_et.ref_emissivity[pos] = model.config.evapotranspiration.sealed_emissivity
                s_et.sealed_interception[pos] = 0.0
            else:
                s_et.ref_albedo[pos] = model.config.evapotranspiration.grass_albedo
                s_et.ref_emissivity[pos] = model.config.evapotranspiration.grass_emissivity

    def _climate_correction(self):
        """
        Calculate climate correction term for the crop coefficients.
        """
        model = self.model
        s = self.model.state
        s_et = s.evapotranspiration

        for lcc, pos in self.land_cover_class_pixels.items():
            plant_height = model.config["land_cover"]["classes"][lcc]["max_height"]
            s_et.clim_corr[pos] = climate_correction(
                model.config.evapotranspiration.mean_wind_speed,
                model.config.evapotranspiration.mean_min_humidity,
                plant_height,
            )

    def evapotranspiration(self):
        model = self.model
        roi = model.grid.roi
        s = model.state
        s_et = s.evapotranspiration
        snowies_roi = s.snow.swe[roi] > 0.0

        s_et.crop_coeff[roi] = np.nan
        s_et.basal_crop_coeff[roi] = np.nan
        s_et.evaporation_coeff[roi] = np.nan
        s_et.water_stress_coeff[roi] = np.nan
        s_et.deep_percolation[roi] = 0.0

        logger.debug("Calculating evapotranspiration")

        self._reference_evapotranspiration()

        for lcc, pos in self.land_cover_class_pixels.items():
            lcc_params = model.config["land_cover"]["classes"][lcc]

            # Derive global masks for pixels with the current land cover class which are
            # snow-covered and snow-free
            pos_snow = model.global_mask(pos[roi] & snowies_roi)
            pos_snowfree = model.global_mask(pos[roi] & (~snowies_roi))

            if lcc_params.get("is_sealed", False):
                # Sealed surfaces are treated separately
                s_et.evaporation[pos_snow] = 0.0
                s_et.transpiration[pos_snow] = 0.0
                s_et.evapotranspiration[pos_snow] = 0.0
                self._sealed_evaporation(pos_snowfree, lcc)
                continue

            crop_coefficient_type = lcc_params["crop_coefficient_type"]
            growing_period_day = model.land_cover.growing_period_day(lcc)
            crop_coeff_ini, crop_coeff_mid, crop_coeff_end = lcc_params["crop_coefficients"]
            is_water_body = lcc_params.get("is_water_body", False)
            (
                length_ini,
                length_dev,
                length_mid,
                length_late,
            ) = model.land_cover.growth_stage_lengths(lcc)

            # Adjust Kcb_mid for sparse vegetation
            if lcc_params.get("is_sparse", False):
                crop_coeff_mid = sparse_vegetation_adjustment(
                    model.config.evapotranspiration.min_crop_coefficient,
                    crop_coeff_mid,
                    lcc_params.sparse_vegetation_fraction,
                    lcc_params.max_height,
                    np.deg2rad(model.grid.center_lat),
                    np.deg2rad(model.sun_params["declination_angle"]),
                )

            # Apply climate correction for Kcb_mid and Kcb_end values >= 0.45 (eq. (70))
            # (convert crop_coeff_mid and crop_coeff_end into fields to allow for possibly
            # non-uniform climate correction values)
            if crop_coeff_mid >= 0.45:
                crop_coeff_mid = np.full(pos.sum(), crop_coeff_mid) + s_et.clim_corr[pos]
            if crop_coeff_end >= 0.45:
                crop_coeff_end = np.full(pos.sum(), crop_coeff_end) + s_et.clim_corr[pos]

            (crop_coeff, plant_height) = crop_coefficient(
                growing_period_day,
                length_ini,
                length_dev,
                length_mid,
                length_late,
                crop_coeff_ini,
                crop_coeff_mid,
                crop_coeff_end,
                model.config.evapotranspiration.min_crop_coefficient,
                max_plant_height=lcc_params.max_height,
            )

            # If scale_height is True, set the plant height to the calculated value according to the
            # crop coefficient curve, otherwise assume a constant height over the season
            if lcc_params.get("scale_height", True):
                s.land_cover.plant_height[pos] = plant_height
            else:
                s.land_cover.plant_height[pos] = lcc_params.max_height

            # Ignore snow cover for water
            if is_water_body:
                pos_snowfree = model.global_mask(pos[roi])
                pos_snow = np.full(pos_snowfree.shape, False)

            # Calculate crop ET under standard conditions
            if crop_coefficient_type == "single":
                s_et.crop_coeff[pos] = crop_coeff
                self._single_coeff_crop_et(pos_snowfree)
                s_et.evapotranspiration[pos_snow] = 0.0
            elif crop_coefficient_type == "dual":
                s_et.basal_crop_coeff[pos] = crop_coeff
                self._dual_coeff_crop_et(pos_snowfree, lcc)
                s_et.evaporation[pos_snow] = 0.0
                s_et.transpiration[pos_snow] = 0.0
                s_et.evapotranspiration[pos_snow] = 0.0
            else:
                raise NotImplementedError

            # Adjust ET for soil water stress conditions (except for water bodies)
            if not is_water_body:
                self._water_stress_coefficient(pos_snowfree)
                if crop_coefficient_type == "single":
                    s_et.evapotranspiration[pos_snowfree] *= s_et.water_stress_coeff[pos_snowfree]  # eq. (81) # fmt: skip # noqa: E501
                elif crop_coefficient_type == "dual":
                    s_et.transpiration[pos_snowfree] *= s_et.water_stress_coeff[pos_snowfree]
                    s_et.evapotranspiration[pos_snowfree] = (
                        s_et.evaporation[pos_snowfree] + s_et.transpiration[pos_snowfree]
                    )
                self._root_zone_water_balance(pos)

    def _reference_net_radiation(self):
        """
        Calculate the reference net radiation, i.e. assuming grass albedo and
        emissivity for non-sealed surfaces and sealed surface albedo and
        emissivity for sealed surfaces.
        """
        model = self.model
        roi = model.grid.roi
        s = model.state
        s_et = s.evapotranspiration

        sw_bal = (1 - s_et.ref_albedo[roi]) * s.meteo.top_canopy_sw_in[roi]
        lw_bal = s.meteo.top_canopy_lw_in[roi] - (
            c.STEFAN_BOLTZMANN * s_et.ref_emissivity[roi] * s.meteo.top_canopy_temp[roi] ** 4
        )
        s_et.ref_net_radiation[roi] = sw_bal + lw_bal

    def _reference_evapotranspiration(self):
        """
        Calculate reference evapotranspiration (ETo).
        """
        model = self.model
        roi = model.grid.roi
        s = model.state
        s_et = s.evapotranspiration

        self._reference_net_radiation()

        soil_heat_flux_factor = 0.1 if model.sun_params["sun_over_horizon"] else 0.5
        s_et.soil_heat_flux[roi] = soil_heat_flux_factor * s_et.ref_net_radiation[roi]  # eq. (45-46) # fmt: skip # noqa: E501

        Wm2_to_MJm2h = (
            1e-6 * c.SECONDS_PER_HOUR
        )  # conversion factor from W m-2 (= J m-2 s-1) to MJ m-2 h-1

        Rn = s_et.ref_net_radiation[roi] * Wm2_to_MJm2h  # net radiation at the grass surface (MJ m-2 h-1) # fmt: skip # noqa: E501
        G = s_et.soil_heat_flux[roi] * Wm2_to_MJm2h  # soil heat flux density (MJ m-2 h-1)
        T = s.meteo.top_canopy_temp[roi] - c.T0  # air temperature (°C)
        D = 4098 * (0.6108 * np.exp(17.27 * T / (T + 237.3))) / (T + 237.3) ** 2  # slope of the relationship between saturation vapor pressure and temperature (kPa °C-1) (eq. (13)) # fmt: skip # noqa: E501
        gamma = s.meteo.psych_const[roi] * 1e-3  # psychrometric constant (kPa °C-1)
        es = s.meteo.sat_vap_press[roi] * 1e-3  # saturation vapor pressure (kPa)
        ea = s.meteo.vap_press[roi] * 1e-3  # actual vapor pressure (kPa)

        grass_roughness_length = 0.03  # (m)
        u2 = meteo.log_wind_profile(  # 2 m wind speed (m s-1)
            s.meteo.top_canopy_wind_speed[roi],
            model.config.meteo.measurement_height.wind,
            2,
            grass_roughness_length,
        )

        ET0 = (  # reference evapotranspiration (kg m-2 h-1) (eq. (53))
            (0.408 * D * (Rn - G) + gamma * 37 / (T + 273) * u2 * (es - ea))
            / (D + gamma * (1 + 0.34 * u2))
        )
        ET0 = ET0.clip(min=0)  # do not allow negative values
        s_et.et_ref[roi] = ET0 * model.timestep / c.SECONDS_PER_HOUR  # (kg m-2)

    def _single_coeff_crop_et(self, pos):
        s_et = self.model.state.evapotranspiration
        et_ref = s_et.et_ref[pos].copy()
        et_ref[np.isnan(et_ref)] = 0.0
        s_et.evapotranspiration[pos] = s_et.crop_coeff[pos] * et_ref

    def _dual_coeff_crop_et(self, pos, lcc):
        model = self.model
        lcc_params = model.config["land_cover"]["classes"][lcc]
        s = model.state
        s_et = s.evapotranspiration
        et_ref = s_et.et_ref[pos].copy()
        et_ref[np.isnan(et_ref)] = 0.0

        plant_height = s.land_cover.plant_height[pos]
        min_crop_coeff = model.config.evapotranspiration.min_crop_coefficient

        # Calculate K_c_max (eq. (72))
        # TODO the climate correction term is intended to be calculated using mean values of wind
        # speed and daily-minimum relative humidity over the period of interest - maybe better use
        # 24-hour moving averages than instantaneous values?
        max_crop_coeff = np.maximum(1.2 + s_et.clim_corr[pos], s_et.basal_crop_coeff[pos] + 0.05)

        # Calculate fraction of the soil surface covered by vegetation. For the special case of
        # sparse vegetation use the fixed defined value, otherwise calculate the vegetation
        # fraction using eq. (76).
        if lcc_params.get("is_sparse", False):
            veg_frac = lcc_params.sparse_vegetation_fraction
        else:
            veg_frac = (
                (s_et.basal_crop_coeff[pos] - min_crop_coeff).clip(min=0.01)
                / (max_crop_coeff - min_crop_coeff)
            ) ** (1 + 0.5 * plant_height)

        # Fraction of soil surface wetted by irrigation or precipitation (use value for
        # precipitation (= 1.0) from Table 20)
        wetted_frac = 1.0

        # Exposed and wetted soil fraction (eq. (75))
        exposed_wetted_frac = np.minimum(1 - veg_frac, wetted_frac)

        # Calculate evaporation reduction coefficient (eq. (74))
        pos2 = s_et.cum_soil_surface_depletion[pos] > s_et.readily_evaporable_water[pos]
        pos3 = model.global_mask(pos2, pos)
        evaporation_reduction_coeff = np.ones(pos.sum())  # K_r = 1 when D_e,i-1 <= REW
        evaporation_reduction_coeff[pos2] = (
            s_et.total_evaporable_water[pos3] - s_et.cum_soil_surface_depletion[pos3]
        ) / (s_et.total_evaporable_water[pos3] - s_et.readily_evaporable_water[pos3])

        # Calculate evaporation coefficient (eq. (71))
        s_et.evaporation_coeff[pos] = np.minimum(
            evaporation_reduction_coeff * (max_crop_coeff - s_et.basal_crop_coeff[pos]),
            exposed_wetted_frac * max_crop_coeff,
        )

        s_et.crop_coeff[pos] = s_et.basal_crop_coeff[pos] + s_et.evaporation_coeff[pos]
        s_et.evaporation[pos] = s_et.evaporation_coeff[pos] * et_ref
        s_et.transpiration[pos] = s_et.basal_crop_coeff[pos] * et_ref
        s_et.evapotranspiration[pos] = s_et.evaporation[pos] + s_et.transpiration[pos]

        # Initialize water balance variables
        precip = np.nan_to_num(s.meteo.rainfall[pos])
        precip_runoff = 0.0  # as suggested by [1]
        irrigation = 0.0
        soil_transpiration = 0.0  # as suggested by [1]

        # Calculate water balance for the surface soil layer
        s_et.deep_percolation_evaporation_layer[pos] = (  # eq. (79)
            precip
            - precip_runoff
            + irrigation / wetted_frac
            - s_et.deep_percolation_evaporation_layer[pos]
        ).clip(min=0)
        s_et.cum_soil_surface_depletion[pos] = (  # eq. (77)
            s_et.cum_soil_surface_depletion[pos]
            - (precip - precip_runoff)
            - irrigation / wetted_frac
            + s_et.evaporation[pos] / exposed_wetted_frac
            + soil_transpiration
            + s_et.deep_percolation_evaporation_layer[pos]
        ).clip(min=0)
        s_et.cum_soil_surface_depletion[pos] = np.minimum(  # eq. (78)
            s_et.cum_soil_surface_depletion[pos],
            s_et.total_evaporable_water[pos],
        )

    def _water_stress_coefficient(self, pos):
        model = self.model
        s = model.state
        s_et = s.evapotranspiration

        s_et.water_stress_coeff[pos] = (  # eq. (84)
            (s_et.total_available_water[pos] - s_et.cum_root_zone_depletion[pos])
            / (s_et.total_available_water[pos] - s_et.readily_available_water[pos])
        ).clip(min=0, max=1)

    def _root_zone_water_balance(self, pos):
        """
        Calculate water balance for the root zone.
        """
        model = self.model
        s = model.state
        s_et = s.evapotranspiration

        precip = np.nan_to_num(s.meteo.rainfall[pos])
        precip_runoff = 0.0  # as suggested by [1]
        irrigation = 0.0
        capillary_rise = 0.0  # assumed to be zero when the water table is more than about 1 m below the bottom of the root zone [1] # noqa: E501

        s_et.deep_percolation[pos] = (  # eq. (88)
            (precip - precip_runoff)
            + irrigation
            - s_et.evapotranspiration[pos]
            - s_et.cum_root_zone_depletion[pos]
        ).clip(min=0)

        s_et.cum_root_zone_depletion[pos] = (  # eq. (85)
            s_et.cum_root_zone_depletion[pos]
            - (precip - precip_runoff)
            - irrigation
            - capillary_rise
            + s_et.evapotranspiration[pos]
            + s_et.deep_percolation[pos]
        ).clip(min=0)
        s_et.cum_root_zone_depletion[pos] = np.minimum(  # eq. (86)
            s_et.cum_root_zone_depletion[pos],
            s_et.total_available_water[pos],
        )

    def _sealed_evaporation(self, pos, lcc):
        """
        Calculate evaporation for sealed surfaces using the Penman-Monteith
        equation.
        """
        model = self.model
        s = model.state
        s_et = s.evapotranspiration
        lcc_params = model.config.land_cover.classes[lcc]
        max_interception = lcc_params.max_sealed_interception

        pos_rain_local = s.meteo.rainfall[pos] > 0.0
        pos_rain = model.global_mask(pos_rain_local, pos)
        pos_dry = model.global_mask(~pos_rain_local, pos)

        s_et.sealed_interception[pos_rain] += np.nan_to_num(s.meteo.rainfall[pos_rain])
        runoff = (s_et.sealed_interception[pos_rain] - max_interception).clip(min=0)
        s_et.sealed_interception[pos_rain] -= runoff
        s_et.deep_percolation[pos_rain] = runoff  # runoff is currently treated as deep percolation for sealed surfaces (should be improved) # fmt: skip # noqa: E501
        s_et.evaporation[pos_rain] = 0.0

        rs = 0.0  # stomatal resistance (s m-1)
        zom = 0.123 * lcc_params.max_height  # roughness length governing momentum transfer (m)
        zoh = 0.1 * zom  # roughness length governing heat and vapor (m)
        d = 2.0 / 3 * lcc_params.max_height  # zero-plane displacement height (m)
        ra = (  # aerodynamic resistance (s m-1) (eq. (4))
            np.log((model.config.meteo.measurement_height.wind - d) / zom)
            * np.log((model.config.meteo.measurement_height.temperature - d) / zoh)
            / (c.VON_KARMAN**2 * s.meteo.top_canopy_wind_speed[pos_dry])
        )

        Rn = s_et.ref_net_radiation[pos_dry]  # net radiation (W m-2)
        G = s_et.soil_heat_flux[pos_dry]  # soil heat flux density (W m-2)
        T = s.meteo.top_canopy_temp[pos_dry] - c.T0  # air temperature (°C)
        D = 1e3 * 4098 * (0.6108 * np.exp(17.27 * T / (T + 237.3))) / (T + 237.3) ** 2  # slope of the relationship between saturation vapor pressure and temperature (Pa K-1) (eq. (13)) # fmt: skip # noqa: E501

        gamma = s.meteo.psych_const[pos_dry]  # psychrometric constant (Pa K-1)
        es = s.meteo.sat_vap_press[pos_dry]  # saturation vapor pressure (Pa)
        ea = s.meteo.vap_press[pos_dry]  # actual vapor pressure (Pa)

        Tv = 1.01 * s.meteo.top_canopy_temp[pos_dry]  # virtual temperature (K)
        rhoa = s.meteo.atmos_press[pos_dry] / c.GAS_CONSTANT_DRY_AIR * Tv  # air density at constant pressure (kg m-3) # fmt: skip # noqa: E501
        cp = (  # specific heat at constant pressure (J kg-1 K-1) (p. 26)
            gamma * 0.622 * c.LATENT_HEAT_OF_VAPORIZATION / s.meteo.atmos_press[pos_dry]
        )

        evaporation_Wm2 = (  # evaporation (W m-2) (eq. (3))
            (D * (Rn - G) + rhoa * cp * (es - ea) / ra) / (D + gamma * (1 + rs / ra))
        ).clip(min=0)
        evaporation_Wm2[np.isnan(evaporation_Wm2)] = 0.0
        s_et.evaporation[pos_dry] = np.minimum(  # (kg m-2)
            evaporation_Wm2 / c.LATENT_HEAT_OF_VAPORIZATION * model.timestep,
            s_et.sealed_interception[pos_dry],
        )
        s_et.sealed_interception[pos_dry] -= s_et.evaporation[pos_dry]

        s_et.transpiration[pos] = 0.0
        s_et.evapotranspiration[pos] = s_et.evaporation[pos] + s_et.transpiration[pos]


def climate_correction(mean_wind_speed, mean_min_rel_hum, plant_height):
    """
    Calculate the climate correction term for the crop coefficients (right part
    of eq. (70) from [1]).

    Parameters
    ----------
    mean_wind_speed : float or ndarray(float)
        Mean wind speed value during the period of interest (m s-1).

    mean_min_rel_hum : float or ndarray(float)
        Mean value for daily minimum relative humidity during the period of
        interest (%).

    plant_height : float
        Plant height during the period of interest (m).

    Returns
    -------
    clim_corr : float or ndarray(float)
        Climate correction term.

    References
    ----------
    .. [1] Allen, R.G., Pereira, L.S., Raes, D., et al. (1998). Crop
       Evapotranspiration-Guidelines for Computing Crop Water Requirements-FAO
       Irrigation and Drainage Paper 56. FAO, Rome, 300(9): D05109.
       http://www.fao.org/3/x0490e/x0490e00.htm
       https://www.researchgate.net/publication/284300773_FAO_Irrigation_and_drainage_paper_No_56
    """
    mean_wind_speed = np.clip(mean_wind_speed, 1, 6)
    mean_min_rel_hum = np.clip(mean_min_rel_hum, 20, 80)
    plant_height = np.clip(plant_height, 0.1, 10)

    return np.nan_to_num(
        (0.04 * (mean_wind_speed - 2) - 0.004 * (mean_min_rel_hum - 45)) * (plant_height / 3) ** 0.3
    )


def crop_coefficient(
    growing_period_day,
    length_ini,
    length_dev,
    length_mid,
    length_late,
    crop_coeff_ini,
    crop_coeff_mid,
    crop_coeff_end,
    crop_coeff_min,
    max_plant_height=None,
):
    """
    Calculate the crop coefficient K_c or basal crop coefficient K_cb for a
    given day following [1].

    The lengths of the individual crop development stages can either be scalars
    or arrays. The former corresponds to calculation of a single crop
    coefficient curve (see [1], p. 127, "Annual crops"), whereas the latter
    corresponds to a crop coefficient curve composed of a series of subcycles
    (see [1], p. 127, "K_c curves for forage crops").

    If the `max_plant_height` parameter is set, in addition to the crop
    coefficient the plant height for the given day is calculated by multiplying
    the maximum plant height by Kcb/Kcb_mid, while assuming that the plant
    height does not decrease with time (see [1], p. 277, footnote 3).

    Parameters
    ----------
    growing_period_day : int
        Day within the growing period (1 = first day of the period).

    length_ini : int or array-like
        Length(s) of the initial growth stage(s) (days).

    length_dev : int or array-like
        Length(s) of the crop development stage(s) (days).

    length_mid : int or array-like
        Length(s) of the mid-season stage(s) (days).

    length_late : int or array-like
        Length(s) of the late season stage(s) (days).

    crop_coeff_ini : float
        Crop coefficient for the initial stage.

    crop_coeff_mid : float or ndarray(float)
        Crop coefficient for the mid-season stage.

    crop_coeff_end : float or ndarray(float)
        Crop coefficient for the end of the late season stage.

    crop_coeff_min : float
        Crop coefficient outside of the growing period.

    max_plant_height : float, default None
        Maximum plant height during the mid-season stage.

    Returns
    -------
    crop_coeff : float or ndarray(float)
        Crop coefficient for the given day.
        Depending on the data types of crop_coeff_mid and crop_coeff_end, this
        is either a scalar or an array.

    plant_height : float or ndarray(float)
        Plant height for the given day (only returned if max_plant_height is
        not None).

    References
    ----------
    .. [1] Allen, R.G., Pereira, L.S., Raes, D., et al. (1998). Crop
       Evapotranspiration-Guidelines for Computing Crop Water Requirements-FAO
       Irrigation and Drainage Paper 56. FAO, Rome, 300(9): D05109.
       http://www.fao.org/3/x0490e/x0490e00.htm
       https://www.researchgate.net/publication/284300773_FAO_Irrigation_and_drainage_paper_No_56
    """
    length_ini = np.atleast_1d(length_ini)
    length_dev = np.atleast_1d(length_dev)
    length_mid = np.atleast_1d(length_mid)
    length_late = np.atleast_1d(length_late)

    if not (length_ini.shape == length_dev.shape == length_mid.shape == length_late.shape):
        raise ValueError("Growth period length arrays have unequal sizes")

    lengths = np.vstack((length_ini, length_dev, length_mid, length_late)).flatten(order="F")
    lengths_cum = lengths.cumsum()
    idx = np.searchsorted(np.concatenate([[0], lengths_cum]), growing_period_day)
    period_num = (idx - 1) % 4

    if idx in (0, len(lengths_cum) + 1):  # outside of growing period
        crop_coeff = crop_coeff_min
    elif period_num == 0:  # initial
        crop_coeff = crop_coeff_ini
    elif period_num == 1:  # crop development
        crop_coeff = (  # eq. (66)
            crop_coeff_ini
            + (growing_period_day - lengths_cum[idx - 2])
            / lengths[idx - 1]
            * (crop_coeff_mid - crop_coeff_ini)
        )
    elif period_num == 2:  # mid season
        crop_coeff = crop_coeff_mid
    elif period_num == 3:  # late season
        crop_coeff = (  # eq. (66)
            crop_coeff_mid
            + (growing_period_day - lengths_cum[idx - 2])
            / lengths[idx - 1]
            * (crop_coeff_end - crop_coeff_mid)
        )

    if max_plant_height is not None:
        if idx in (0, len(lengths_cum) + 1):  # outside of growing period
            min_plant_height = 0.0
        elif period_num == 0:  # initial
            min_plant_height = crop_coeff_min / crop_coeff_mid * max_plant_height
        elif period_num == 1:  # crop development
            min_plant_height = crop_coeff_ini / crop_coeff_mid * max_plant_height
        elif period_num == 2 or period_num == 3:  # mid season
            min_plant_height = max_plant_height

        plant_height = np.maximum(
            crop_coeff / crop_coeff_mid * max_plant_height,
            min_plant_height,
        )

    if max_plant_height is None:
        return crop_coeff
    else:
        return (crop_coeff, plant_height)


def sparse_vegetation_adjustment(
    crop_coeff_min,
    crop_coeff_mid,
    veg_frac,
    plant_height,
    lat,
    declination_angle,
):
    """
    Adjust the mid-season crop coefficient for sparsely covered vegetation
    following [1] (Chapter 9), assuming round or spherical shaped canopies
    (such as trees).

    Parameters
    ----------
    crop_coeff_min : float
        Crop coefficient outside of the growing period.

    crop_coeff_mid : float or ndarray(float)
        Unadjusted crop coefficient for the mid-season stage.

    veg_frac : float
        Fraction of soil surface covered by vegetation.

    plant_height : float
        Plant height (m).

    lat : float
        Latitude (radians).

    declination_angle : float
        Solar declination angle (radians).

    Returns
    -------
    crop_coeff_mid_adj : float or ndarray(float)
        Adjusted mid-season crop coefficient.

    References
    ----------
    .. [1] Allen, R.G., Pereira, L.S., Raes, D., et al. (1998). Crop
       Evapotranspiration-Guidelines for Computing Crop Water Requirements-FAO
       Irrigation and Drainage Paper 56. FAO, Rome, 300(9): D05109.
       http://www.fao.org/3/x0490e/x0490e00.htm
       https://www.researchgate.net/publication/284300773_FAO_Irrigation_and_drainage_paper_No_56
    """
    # Mean angle above the sun during the period of maximum evapotranspiration
    time_angle = 0.0  # calculate for solar noon (12:00), i.e., time angle = 0
    sin_mean_angle_above_sun = np.sin(lat) * np.sin(declination_angle) + np.cos(lat) * np.cos(
        declination_angle
    ) * np.cos(time_angle)

    veg_frac_eff = np.array(veg_frac / sin_mean_angle_above_sun).clip(min=0)
    crop_coeff_mid_adj = (  # eq. (98)
        crop_coeff_min
        + (crop_coeff_mid - crop_coeff_min)
        * min(
            1,
            2 * veg_frac,
            veg_frac_eff ** (1 / (1 + plant_height)),
        )
    )
    return crop_coeff_mid_adj



import warnings

import numpy as np

from openamundsen import constants as c
from openamundsen import meteo


class CanopyModel:
    """
    Canopy model for adjusting meteorology for inside-canopy conditions and
    calculating snow-canopy interactions following [1].

    References
    ----------
    .. [1] Strasser, U., Warscher, M., & Liston, G. E. (2011). Modeling
       Snow-Canopy Processes on an Idealized Mountain. Journal of Hydrometeorology,
       12(4), 663–677.  https://doi.org/10.1175/2011JHM1344.1
    """

    def __init__(self, model):
        self.model = model
        num_timesteps_per_day = int(c.HOURS_PER_DAY * c.SECONDS_PER_HOUR / model.timestep)
        model.state.meteo.add_variable(
            "last_24h_temps",
            "K",
            "Air temperatures of the previous 24 hours",
            dim3=max(num_timesteps_per_day, 1),
            retain=True,
        )
        self._temp_idx = 0  # current index within the last_24h_temps array

        model.state.snow.add_variable(
            "canopy_intercepted_load",
            "kg m-2",
            "Canopy snow interception storage",
            retain=True,
        )
        model.state.snow.add_variable(
            "canopy_intercepted_snowfall",
            "kg m-2",
            "Canopy-intercepted snowfall",
        )
        model.state.snow.add_variable(
            "canopy_sublimation",
            "kg m-2",
            "Canopy snow sublimation",
        )
        model.state.snow.add_variable(
            "canopy_melt",
            "kg m-2",
            "Melt of canopy-intercepted snow",
        )

    def initialize(self):
        model = self.model

        if len(model.land_cover.forest_classes) > 0:
            self.forest_pos = np.logical_or.reduce(
                [model.land_cover.class_pixels[lcc] for lcc in model.land_cover.forest_classes]
            )
        else:
            self.forest_pos = np.full(model.grid.shape, False)

        model.state.snow.canopy_intercepted_load[self.forest_pos] = 0.0

    def meteorology(self):
        """
        Modify meteorology for inside-canopy conditions.
        """
        model = self.model
        s = model.state
        pos = self.forest_pos
        lai_eff = s.land_cover.lai_eff[pos]

        canopy_frac = (0.55 + 0.29 * np.log(lai_eff)).clip(0, 1)  # eq. (2)

        s.meteo.sw_in[pos] *= np.exp(
            -model.config.canopy.extinction_coefficient * lai_eff
        )  # eq. (1)

        s.meteo.last_24h_temps[self._temp_idx, :, :] = s.meteo.temp
        self._temp_idx += 1
        if self._temp_idx == s.meteo.last_24h_temps.shape[0]:
            self._temp_idx = 0

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "Mean of empty slice", RuntimeWarning)
            mean_24h_temp = np.nanmean(s.meteo.last_24h_temps, axis=0)

        pos_finite_temp = pos & np.isfinite(mean_24h_temp)
        delta_t = ((mean_24h_temp[pos_finite_temp] - c.T0) / 3).clip(-2, 2)  # eq. (5)
        s.meteo.temp[pos_finite_temp] -= (  # eq. (4)
            canopy_frac
            * (
                s.meteo.temp[pos_finite_temp]
                - (
                    model.config.canopy.temperature_scaling_coefficient
                    * (s.meteo.temp[pos_finite_temp] - mean_24h_temp[pos_finite_temp])
                    + mean_24h_temp[pos_finite_temp]
                    - delta_t
                )
            )
        )

        s.meteo.lw_in[pos] = (  # eq. (3)
            (1 - canopy_frac) * s.meteo.lw_in[pos]
            + canopy_frac * c.STEFAN_BOLTZMANN * s.meteo.temp[pos] ** 4
        )

        s.meteo.rel_hum[pos] = (  # eq. (6)
            s.meteo.rel_hum[pos] * (1 + 0.1 * canopy_frac)
        ).clip(0, 100)

        # Wind speed is modified assuming a reference level of z = 0.6h, i.e., the term (1 - z/h)
        # from Liston & Elder (2006, eq. (16)) becomes 0.4
        canopy_flow_index = model.config.canopy.canopy_flow_index_coefficient * lai_eff  # eq. (8)
        s.meteo.wind_speed[pos] *= np.exp(-0.4 * canopy_flow_index)  # eq. (7)

    def snow(self):
        """
        Calculate canopy snow processes (interception, sublimation and melt unload).
        """
        model = self.model
        s = model.state
        pos = self.forest_pos

        # Absorbed solar radiation by a snow particle in the canopy (W) (eq. (11))
        absorbed_rad = (
            np.pi
            * model.config.canopy.spherical_ice_particle_radius**2
            * (1 - s.surface.albedo[pos])
            * s.meteo.top_canopy_sw_in[pos]
        )

        reynolds = (  # Reynolds number (eq. (12))
            2
            * model.config.canopy.spherical_ice_particle_radius
            * s.meteo.wind_speed[pos]
            / model.config.canopy.kinematic_air_viscosity
        ).clip(0.7, 10)
        nusselt = 1.79 + 0.606 * np.sqrt(reynolds)  # Nusselt number (eq. (13))
        sherwood = nusselt  # Sherwood number

        # Saturation density of water vapor (kg m-3) (eq. (15))
        wat_vap_sat_dens = 0.622 * meteo.absolute_humidity(
            s.meteo.temp[pos],
            s.meteo.sat_vap_press[pos],
        )

        # Diffusivity of water vapor in the atmosphere (m2 s-1) (eq. (16))
        diff_wat_vap = 2.06e-5 * (s.meteo.temp[pos] / 273.0) ** 1.75

        omega = (  # eq. (18)
            1.0
            / (c.THERM_COND_AIR * s.meteo.temp[pos] * nusselt)
            * (
                c.LATENT_HEAT_OF_SUBLIMATION
                * c.MOLAR_MASS_WATER
                / (c.UNIVERSAL_GAS_CONSTANT * s.meteo.temp[pos])
                - 1.0
            )
        )

        # Mass loss rate from an ice sphere (kg s-1) (eq. (17))
        mass_loss_rate = (
            2
            * np.pi
            * model.config.canopy.spherical_ice_particle_radius
            * (s.meteo.rel_hum[pos] / 100.0 - 1.0)
            - absorbed_rad * omega
        ) / (
            c.LATENT_HEAT_OF_SUBLIMATION * omega
            + 1.0 / (diff_wat_vap * wat_vap_sat_dens * sherwood)
        )
        mass_loss_rate[np.isnan(mass_loss_rate)] = 0.0

        # Particle mass (kg) (eq. (20))
        particle_mass = (
            4.0 / 3.0 * np.pi * c.ICE_DENSITY * model.config.canopy.spherical_ice_particle_radius**3
        )

        # Sublimation loss rate coefficient from an ice sphere (s-1) (eq. (19))
        sublim_loss_rate_coeff = mass_loss_rate / particle_mass

        # Update canopy-intercepted load (kg m-2) (eq. (20))
        max_interception_storage = (
            model.config.canopy.max_interception_storage_coefficient * s.land_cover.lai_eff[pos]
        )
        new_canopy_intercepted_load = s.snow.canopy_intercepted_load[pos] + 0.7 * (
            max_interception_storage - s.snow.canopy_intercepted_load[pos]
        ) * (1 - np.exp(-s.meteo.snowfall[pos] / max_interception_storage))
        s.snow.canopy_intercepted_snowfall[pos] = (
            new_canopy_intercepted_load - s.snow.canopy_intercepted_load[pos]
        ).clip(min=0)
        s.snow.canopy_intercepted_load[pos] = new_canopy_intercepted_load

        # Canopy exposure coefficient (eq. (23))
        exposure_coeff = np.zeros(new_canopy_intercepted_load.shape)
        pos_int = new_canopy_intercepted_load > 0
        exposure_coeff[pos_int] = model.config.canopy.exposure_coefficient_coefficient * (
            new_canopy_intercepted_load[pos_int] / max_interception_storage[pos_int]
        ) ** (-0.4)

        # Canopy sublimation (eq. (22))
        s.snow.canopy_sublimation[pos] = np.minimum(
            (
                -exposure_coeff
                * new_canopy_intercepted_load
                * sublim_loss_rate_coeff
                * model.timestep
            ),
            new_canopy_intercepted_load,
        ).clip(min=0)
        s.snow.canopy_intercepted_load[pos] -= s.snow.canopy_sublimation[pos]

        # Calculate melt unload using a temperature index method
        # Instead of applying an additional scaling factor to the default degree day factors for
        # snow on the ground as in [1], here a distinct degree day factor for canopy-intercepted
        # snow is used directly.
        s.snow.canopy_melt[pos] = np.minimum(
            (
                model.config.canopy.degree_day_factor
                * (s.meteo.temp[pos] - c.T0).clip(min=0)
                * model.timestep
                / (c.SECONDS_PER_HOUR * c.HOURS_PER_DAY)
            ),
            s.snow.canopy_intercepted_load[pos],
        )
        s.snow.canopy_intercepted_load[pos] -= s.snow.canopy_melt[pos]

        # Precipitation reaching the ground: reduce snowfall by canopy interception and increase by
        # melt unload
        s.meteo.snowfall[pos] -= s.snow.canopy_intercepted_snowfall[pos]
        s.meteo.snowfall[pos] += s.snow.canopy_melt[pos]
        s.meteo.precip[pos] = s.meteo.snowfall[pos] + s.meteo.rainfall[pos]


def above_canopy_meteorology(model):
    """
    Save above-canopy meteorological variables before they (potentially)
    are modified by the canopy model.
    """
    s_m = model.state.meteo
    roi = model.grid.roi

    s_m.top_canopy_temp[roi] = s_m.temp[roi]
    s_m.top_canopy_rel_hum[roi] = s_m.rel_hum[roi]
    s_m.top_canopy_wind_speed[roi] = s_m.wind_speed[roi]
    s_m.top_canopy_sw_in[roi] = s_m.sw_in[roi]
    s_m.top_canopy_lw_in[roi] = s_m.lw_in[roi]



from pathlib import Path

import numpy as np
import pytest
from numpy.testing import assert_allclose

import openamundsen as oa

from .conftest import base_config


@pytest.mark.slow
def test_evapotranspiration(tmp_path):
    config = base_config()
    config.start_date = "2020-07-01"
    config.end_date = "2020-07-15"

    model = oa.OpenAmundsen(config)
    model.initialize()

    meteo = model.meteo.copy()
    meteo.temp.values[:, 30] = np.nan  # nan values should not propagate to evapotranspiration variables # fmt: skip # noqa: E501

    roi_xs = model.grid.X.flat[model.grid.roi_idxs_flat]
    roi_ys = model.grid.Y.flat[model.grid.roi_idxs_flat]

    for p in Path(config.input_data.grids.dir).glob("*.asc"):
        (tmp_path / p.name).symlink_to(p)

    soil = np.zeros(model.grid.shape, dtype=int)
    lc = np.zeros(model.grid.shape, dtype=int)

    # Test with all land cover classes and fixed soil texture
    soil[:] = 5
    lccs = model.config.land_cover.classes.keys()
    config.output_data.timeseries.points = []
    for lcc_num, lcc in enumerate(lccs):
        config.output_data.timeseries.points.append(
            {
                "x": float(roi_xs[lcc_num]),
                "y": float(roi_ys[lcc_num]),
            }
        )
        lc.flat[model.grid.roi_idxs_flat[lcc_num]] = lcc

    config.input_data.grids.dir = str(tmp_path)

    rio_meta = {"driver": "AAIGrid"}
    oa.fileio.write_raster_file(
        oa.util.raster_filename("soil", config),
        soil.astype(np.int32),
        model.grid.transform,
        **rio_meta,
    )
    oa.fileio.write_raster_file(
        oa.util.raster_filename("lc", config),
        lc.astype(np.int32),
        model.grid.transform,
        **rio_meta,
    )

    config.evapotranspiration.enabled = True
    config.output_data.timeseries.add_default_points = False
    config.output_data.timeseries.variables = [
        {"var": "evapotranspiration.evaporation"},
        {"var": "evapotranspiration.transpiration"},
        {"var": "evapotranspiration.evapotranspiration"},
    ]

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.meteo = meteo
    model.run()
    ds = model.point_output.data

    for lcc_num, lcc in enumerate(lccs):
        lcc_params = model.config.land_cover.classes[lcc]
        crop_coeff_type = lcc_params.get("crop_coefficient_type", None)
        is_sealed = lcc_params.get("is_sealed", False)

        ds_lcc = ds.isel(point=lcc_num)
        assert np.all(ds_lcc.evapotranspiration >= 0)
        assert ds_lcc.evapotranspiration.max() > 0

        if crop_coeff_type == "dual" or is_sealed:
            assert np.all(ds_lcc.evaporation >= 0)
            assert np.all(ds_lcc.transpiration >= 0)
            assert_allclose(
                ds_lcc.evaporation + ds_lcc.transpiration,
                ds_lcc.evapotranspiration,
                rtol=1e-3,
            )
        elif crop_coeff_type == "single":
            assert np.all(np.isnan(ds_lcc.evaporation))
            assert np.all(np.isnan(ds_lcc.transpiration))

    # Test with all soil texture classes and fixed land cover
    lc[:] = 9
    soil[:] = 0
    stcs = range(1, 9 + 1)

    config.output_data.timeseries.points = []
    for stc_num, stc in enumerate(stcs):
        config.output_data.timeseries.points.append(
            {
                "x": float(roi_xs[stc_num]),
                "y": float(roi_ys[stc_num]),
            }
        )
        soil.flat[model.grid.roi_idxs_flat[stc_num]] = stc

    rio_meta = {"driver": "AAIGrid"}
    oa.fileio.write_raster_file(
        oa.util.raster_filename("soil", config),
        soil.astype(np.int32),
        model.grid.transform,
        **rio_meta,
    )
    oa.fileio.write_raster_file(
        oa.util.raster_filename("lc", config),
        lc.astype(np.int32),
        model.grid.transform,
        **rio_meta,
    )

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.meteo = meteo
    model.run()
    ds = model.point_output.data

    for stc_num in range(len(stcs)):
        ds_stc = ds.isel(point=stc_num)
        assert np.all(ds_stc.evapotranspiration >= 0)
        assert ds_lcc.evapotranspiration.max() > 0
        assert np.all(ds_stc.evaporation >= 0)
        assert np.all(ds_stc.transpiration >= 0)
        assert_allclose(
            ds_stc.evaporation + ds_stc.transpiration,
            ds_stc.evapotranspiration,
            rtol=1e-3,
        )



import numpy as np
import pandas as pd
import pytest
import xarray as xr
from numpy.testing import assert_allclose, assert_array_equal

import openamundsen as oa
import openamundsen.errors as errors

from .conftest import base_config


@pytest.mark.parametrize("fmt", ["netcdf", "csv", "memory"])
def test_formats(fmt, tmp_path):
    config = base_config()
    config.end_date = "2020-01-16"
    config.results_dir = tmp_path
    config.output_data.timeseries.format = fmt
    config.output_data.timeseries.variables = [{"var": "snow.num_layers"}]

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    point_ids = ["bellavista", "latschbloder", "proviantdepot"]

    if fmt in ("netcdf", "memory"):
        if fmt == "netcdf":
            # Set mask_and_scale=False because otherwise integer variables with a _FillValue are
            # converted to float by xarray (see e.g. https://github.com/pydata/xarray/issues/1194),
            # even when they do not actually contain any missing values.
            ds = xr.open_dataset(
                tmp_path / "output_timeseries.nc",
                mask_and_scale=False,
            )
        elif fmt == "memory":
            ds = model.point_output.data

        assert ds.time.to_index().equals(model.dates)
        assert_array_equal(ds.point, point_ids)
        assert_array_equal(
            list(ds.coords.keys()),
            ["time", "point", "lon", "lat", "alt", "x", "y", "soil_layer", "snow_layer"],
        )
        assert ds.temp.dims == ("time", "point")
        assert ds.snow_thickness.dims == ("time", "snow_layer", "point")
        assert ds.soil_temp.dims == ("time", "soil_layer", "point")
        assert ds.temp.dtype == np.float32
        assert np.issubdtype(ds.num_layers.dtype, np.integer)
        assert np.all(ds.temp > 250.0)
    elif fmt == "csv":
        for point_id in point_ids:
            assert (tmp_path / f"point_{point_id}.csv").exists()

        df = pd.read_csv(tmp_path / "point_bellavista.csv", index_col="time", parse_dates=True)
        assert df.index.equals(model.dates)
        assert df.temp.dtype == np.float64
        assert np.issubdtype(df.num_layers.dtype, np.integer)
        assert "snow_thickness0" in df
        assert np.all(df.temp > 250.0)


def test_values():
    config = base_config()
    config.end_date = "2020-01-15"

    model = oa.OpenAmundsen(config)
    model.initialize()

    point = "proviantdepot"
    row = int(model.meteo.sel(station=point).row)
    col = int(model.meteo.sel(station=point).col)

    data_temp = pd.Series(index=model.dates, dtype=float)
    data_soil_temp1 = pd.Series(index=model.dates, dtype=float)

    for date in model.dates:
        model.run_single()
        data_temp[date] = model.state.meteo.temp[row, col]
        data_soil_temp1[date] = model.state.soil.temp[1, row, col]

    ds = model.point_output.data.sel(point=point)
    assert_allclose(ds.temp.values, data_temp)
    assert_allclose(ds.soil_temp.isel(soil_layer=1).values, data_soil_temp1)


@pytest.mark.parametrize("write_freq", ["ME", "7h", "3h", "10min"])
def test_write_freq(write_freq, tmp_path):
    config = base_config()
    config.end_date = "2020-01-15"
    config.results_dir = tmp_path
    config.output_data.timeseries.format = "netcdf"
    config.output_data.timeseries.write_freq = write_freq

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()

    ds = xr.open_dataset(tmp_path / "output_timeseries.nc")
    assert ds.time.to_index().equals(model.dates)


def test_points():
    bc = base_config()
    bc.end_date = "2020-01-15 00:00"

    config = bc.copy()
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert_array_equal(ds.point, ["bellavista", "latschbloder", "proviantdepot"])

    config = bc.copy()
    config.output_data.timeseries.add_default_points = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert ds.point.size == 0

    config = bc.copy()
    config.output_data.timeseries.add_default_points = False
    config.output_data.timeseries.points.append(
        {
            "x": 640367,
            "y": 5182896,
        }
    )
    config.output_data.timeseries.points.append(
        {
            "x": 645378,
            "y": 5190907,
            "name": "mypoint",
        }
    )
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert_array_equal(ds.point, ["point1", "mypoint"])
    assert_allclose(ds.alt, [3181.89, 1948.97])

    # Duplicate point name
    config = bc.copy()
    config.output_data.timeseries.points.append(
        {
            "x": 640367,
            "y": 5182896,
            "name": "bellavista",
        }
    )
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()

    # Duplicate point name
    config = bc.copy()
    config.output_data.timeseries.points.append(
        {
            "x": 640367,
            "y": 5182896,
            "name": "bellavista",
        }
    )
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()

    # Point not within grid
    config = bc.copy()
    config.output_data.timeseries.points.append(
        {
            "x": 637152,
            "y": 5196427,
        }
    )
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()


def test_variables():
    bc = base_config()
    bc.end_date = "2020-01-15 00:00"
    bc.output_data.timeseries.variables = []

    config = bc.copy()
    config.output_data.timeseries.add_default_variables = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert len(ds.data_vars) == 0

    config = bc.copy()
    config.output_data.timeseries.add_default_variables = False
    config.output_data.timeseries.variables.append({"var": "meteo.spec_heat_cap_moist_air"})
    config.output_data.timeseries.variables.append(
        {
            "var": "surface.conductance",
            "name": "myvar",
        }
    )
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds = model.point_output.data
    assert_array_equal(ds.data_vars, ["spec_heat_cap_moist_air", "myvar"])

    # Invalid variable name
    config = bc.copy()
    config.output_data.timeseries.variables.append({"var": "meteo.asdf"})
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()

    # Output name already in use
    config = bc.copy()
    config.output_data.timeseries.variables.append({"var": "meteo.temp"})
    model = oa.OpenAmundsen(config)
    with pytest.raises(errors.ConfigurationError):
        model.initialize()



import datetime
import json
import re
from pathlib import Path

import cerberus
import pandas as pd
import pyproj
from munch import Munch

from openamundsen import constants, util
from openamundsen.errors import ConfigurationError

try:
    import openamundsen_snowmanagement

    SNOW_MANAGEMENT_AVAILABLE = True
    SNOW_MANAGEMENT_DATA_DIR = Path(openamundsen_snowmanagement.__file__).parent / "data"
except ImportError:
    SNOW_MANAGEMENT_AVAILABLE = False

DATA_DIR = Path(__file__).parent / "data"


class ConfigurationValidator(cerberus.Validator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _normalize_coerce_float(self, value):
        return float(value)

    def _normalize_coerce_datetime(self, date):
        return pd.to_datetime(date)

    def _normalize_coerce_path(self, path):
        return Path(path)

    def _normalize_coerce_timestep(self, timestep):
        # As of pandas 2.2.0, some frequency aliases have been deprecated (see
        # https://pandas.pydata.org/docs/dev/whatsnew/v2.2.0.html). Since especially the "H" alias
        # (which is now deprecated in favor of "h") has been frequently used in earlier versions of
        # openAMUNDSEN, we replace it here manually with its lowercase version in order to avoid
        # deprecation warnings.
        if timestep.endswith("H"):
            timestep = timestep[:-1] + "h"
        return timestep

    def _normalize_coerce_snowmodel(self, model):
        # Snow model name 'layers' is deprecated -> change to 'multilayer'
        if model == "layers":
            return "multilayer"

        return model

    def _normalize_coerce_meteo_interpolation(self, d):
        # 'wind_speed' is deprecated -> change to 'wind'
        if "wind_speed" in d:
            d["wind"] = d.pop("wind_speed")

        return d

    def _normalize_coerce_cloudiness(self, d):
        # 'day_method' is deprecated -> change to 'method'
        if "day_method" in d:
            if "method" not in d:
                d["method"] = d.pop("day_method")
            else:
                raise ConfigurationError(
                    "cloudiness.day_method is deprecated. Please use cloudiness.method instead."
                )

        # 'night_method' is deprecated -> change to 'clear_sky_fraction_night_method'
        if "night_method" in d:
            d["clear_sky_fraction_night_method"] = d.pop("night_method")

        return d


class ConfigurationEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, pd.Timestamp):
            return str(o)
        else:
            raise TypeError


def _parse_end_date(end_date, timestep):
    # If end_date is specified without an hour value, the end hour should be inferred
    # (i.e., set to the latest time step of the end day).
    if (
        type(end_date)
        is datetime.date  # do not use isinstance() because we want to catch only datetime.date and not its subclasses (datetime.datetime, pd.Timestamp etc.) # noqa: E501
        or (isinstance(end_date, str) and re.match(r"^\d\d\d\d-\d\d-\d\d$", end_date.strip()))
    ):
        infer_end_hour = True
    else:
        infer_end_hour = False

    end_date = pd.to_datetime(end_date)

    # If no end hour is specified (only the date), set it to the last time step
    # of the respective day (for the start date the hour is automatically set
    # to 0 if not explicitly specified)
    timedelta = util.offset_to_timedelta(timestep)
    if infer_end_hour and timedelta < pd.Timedelta(days=1):
        end_date += pd.Timedelta(hours=24) - timedelta

    return end_date


class Configuration(Munch):
    """
    Container for storing model configuration. This class inherits from `Munch`
    so that attributes are accessible both using dict notation
    (`config['start_date']`) as well as dot notation (`config.end_date`).
    """

    @classmethod
    def from_dict(cls, d):
        return cls.fromDict(d)

    @classmethod
    def from_yaml(cls, s):
        return cls.from_dict(util.load_yaml(s))

    def __repr__(self):
        return util.to_yaml(self.toDict())


def read_config(filename):
    """
    Read a configuration (YAML) file and return the resulting dict as a
    Configuration object.
    """
    return Configuration.fromDict(util.read_yaml_file(filename))


def parse_config(config):
    _init_schemas()
    v = ConfigurationValidator(cerberus.schema_registry.get("openamundsen_config"))
    valid = v.validate(config)

    if not valid:
        raise ConfigurationError("Invalid configuration\n\n" + util.to_yaml(v.errors))

    full_config = Configuration.fromDict(v.document)
    full_config["end_date"] = _parse_end_date(full_config["end_date"], full_config["timestep"])

    full_config["land_cover"]["classes"] = _merge_land_cover_params(
        full_config["land_cover"]["classes"]
    )

    validate_config(full_config)

    return full_config


def _init_schemas():
    """
    Read in the validation schemas and add them to the Cerberus registry.
    """
    schema_reg = cerberus.schema_registry
    schemas = schema_reg.all()

    if "openamundsen_config" not in schemas:
        config_schema = util.read_yaml_file(f"{DATA_DIR}/configschema.yml")

        if SNOW_MANAGEMENT_AVAILABLE:
            sm_schema = util.read_yaml_file(f"{SNOW_MANAGEMENT_DATA_DIR}/configschema.yml")
            config_schema.update(sm_schema)

        schema_reg.add("openamundsen_config", config_schema)

        lcc_schema = util.read_yaml_file(f"{DATA_DIR}/land_cover_class_schema.yml")
        schema_reg.add("openamundsen_land_cover_class", lcc_schema)


def _merge_land_cover_params(class_params):
    """
    Merge the manually set land cover class parameters with the default parameters.
    """
    default_config = read_config(f"{DATA_DIR}/land_cover_class_params.yml")

    dict_schema = {
        "classes": {
            "type": "dict",
            "valuesrules": {
                "schema": "openamundsen_land_cover_class",
            },
        },
    }
    v = cerberus.Validator(dict_schema)
    valid = v.validate(default_config)

    if not valid:
        raise ConfigurationError("Invalid land cover configuration\n\n" + util.to_yaml(v.errors))

    default_class_params = Configuration.fromDict(v.document)["classes"]
    merged_params = default_class_params.copy()

    for lcc, lcc_params in class_params.items():
        if lcc in merged_params:
            for k, v in lcc_params.items():
                merged_params[lcc][k] = v
        else:
            merged_params[lcc] = lcc_params

    return merged_params


def validate_config(config):
    """
    Perform some additional validations which are too complicated with Cerberus.
    """
    if config.start_date > config.end_date:
        raise ConfigurationError("End date must be after start date")

    # Check if timestep matches start/end dates
    dates = pd.date_range(
        start=config.start_date,
        end=config.end_date,
        freq=config.timestep,
    )
    if dates[-1] != config.end_date:
        raise ConfigurationError("Start/end date is not compatible with timestep")

    # Check if a projected CRS is used
    if pyproj.CRS(config.crs).is_geographic:
        raise ConfigurationError(
            f"{config.crs} is a geographic CRS, please use a projected CRS instead"
        )

    # Check if write_freq is compatible with timestep - as long as the time step is <= 1d,
    # write_freq can be an offset like 'ME' or 'YE', but for larger timesteps it is not guaranteed
    # that these dates generated with pd.date_range(start=start_date, end=end_date, freq=write_freq)
    # are actually reached, so in this case write_freq must be a multiple of timestep
    # (e.g. timestep = '5D' and write_freq = '30D')
    timestep_td = util.offset_to_timedelta(config.timestep)
    write_freq = config.output_data.timeseries.write_freq
    if timestep_td > pd.Timedelta(days=1):
        try:
            write_freq_td = pd.Timedelta(write_freq)
            if write_freq_td.total_seconds() % timestep_td.total_seconds() != 0:
                raise ConfigurationError("write_freq must be a multiple of timestep")
        except ValueError as err:
            raise ConfigurationError("write_freq must be a multiple of timestep") from err

    if config.input_data.meteo.format == "netcdf" and config.input_data.meteo.crs is not None:
        print(
            "Warning: Ignoring CRS specification for meteo input data "
            "(CRS not required when using NetCDF format)"
        )

    if config.input_data.meteo.format == "memory":
        if (
            config.input_data.meteo.bounds != "grid"
            or len(config.input_data.meteo.exclude) > 0
            or len(config.input_data.meteo.include) > 0
        ):
            print(
                'Warning: "bounds", "exclude" and "include" are currently ignored for '
                'format "memory"'
            )

    if config.snow.model == "multilayer" and config.snow.melt.method != "energy_balance":
        raise ConfigurationError(
            f'Melt method "{config.snow.melt.method}" not supported for the '
            f'snow model "{config.snow.model}"'
        )

    if config.snow.melt.method == "temperature_index":
        if config.snow.melt.degree_day_factor is None:
            raise ConfigurationError("Missing field: snow.melt.degree_day_factor")
    elif config.snow.melt.method == "enhanced_temperature_index":
        if config.snow.melt.degree_day_factor is None:
            raise ConfigurationError("Missing field: snow.melt.degree_day_factor")
        if config.snow.melt.albedo_factor is None:
            raise ConfigurationError("Missing field: snow.melt.albedo_factor")

    if abs(config.meteo.precipitation_phase.threshold_temp) < 20:
        print(
            "Warning: precipitation phase threshold temperature seems to be in °C, converting to K"
        )
        config.meteo.precipitation_phase.threshold_temp += constants.T0

    ac = config.snow.albedo
    if ac.method == "usaco":
        print('Warning: albedo method "usaco" is deprecated, please use "snow_age"')
        ac.method = "snow_age"
        ac.cold_snow_decay_timescale = 1 / ac.k_neg * constants.HOURS_PER_DAY
        ac.melting_snow_decay_timescale = 1 / ac.k_pos * constants.HOURS_PER_DAY
        ac.decay_timescale_determination_temperature = "air"
        ac.refresh_snowfall = ac.significant_snowfall
        ac.refresh_method = "binary"
    elif ac.method == "fsm":
        print('Warning: albedo method "fsm" is deprecated, please use "snow_age"')
        ac.method = "snow_age"
        ac.decay_timescale_determination_temperature = "surface"
        ac.refresh_method = "continuous"

    if (
        config.glaciers.enabled
        and config.glaciers.model == "delta_h"
        and config.snow.model != "cryolayers"
    ):
        raise ConfigurationError(
            'Glacier model "delta_h" can currently only be used with snow model "cryolayers"'
        )

    if SNOW_MANAGEMENT_AVAILABLE:
        openamundsen_snowmanagement.validate_config(config)



import tempfile
from pathlib import Path

import numpy as np
import pytest

import openamundsen as oa

from .compare import compare_datasets
from .conftest import base_config


@pytest.fixture(scope="session")
def canopy_run():
    config = base_config()
    config.start_date = "2019-10-01"
    config.end_date = "2020-05-31"

    model = oa.OpenAmundsen(config)
    model.initialize()

    roi_xs = model.grid.X.flat[model.grid.roi_idxs_flat]
    roi_ys = model.grid.Y.flat[model.grid.roi_idxs_flat]

    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_path = Path(temp_dir)

        for p in Path(config.input_data.grids.dir).glob("*.asc"):
            (tmp_path / p.name).symlink_to(p)

        config.input_data.grids.dir = str(temp_dir)

        soil = np.zeros(model.grid.shape, dtype=int)
        lc = np.zeros(model.grid.shape, dtype=int)

        soil[:] = 5
        lccs = model.config.land_cover.classes.keys()
        config.output_data.timeseries.points = []
        for lcc_num, lcc in enumerate(lccs):
            config.output_data.timeseries.points.append(
                {
                    "x": float(roi_xs[lcc_num]),
                    "y": float(roi_ys[lcc_num]),
                }
            )
            lc.flat[model.grid.roi_idxs_flat[lcc_num]] = lcc

        rio_meta = {"driver": "AAIGrid"}
        oa.fileio.write_raster_file(
            oa.util.raster_filename("soil", config),
            soil.astype(np.int32),
            model.grid.transform,
            **rio_meta,
        )
        oa.fileio.write_raster_file(
            oa.util.raster_filename("lc", config),
            lc.astype(np.int32),
            model.grid.transform,
            **rio_meta,
        )

        config.canopy.enabled = True
        config.output_data.timeseries.add_default_points = False
        config.output_data.timeseries.variables = [
            {"var": "snow.canopy_intercepted_load"},
            {"var": "snow.canopy_intercepted_snowfall"},
            {"var": "snow.canopy_sublimation"},
            {"var": "snow.canopy_melt"},
            {"var": "meteo.top_canopy_temp"},
        ]

        model = oa.OpenAmundsen(config)
        model.initialize()
        model.run()

    return model


@pytest.mark.slow
def test_canopy_snow(canopy_run):
    model = canopy_run
    ds = model.point_output.data
    lccs = model.config.land_cover.classes.keys()

    for lcc_num, lcc in enumerate(lccs):
        lcc_params = model.config.land_cover.classes[lcc]
        is_forest = lcc_params.get("is_forest", False)

        ds_lcc = ds.isel(point=lcc_num)

        if is_forest:
            assert np.all(ds_lcc.canopy_intercepted_load >= 0.0)
            assert np.all(ds_lcc.canopy_intercepted_snowfall >= 0.0)
            assert np.all(ds_lcc.canopy_sublimation >= 0.0)
            assert np.all(ds_lcc.canopy_melt >= 0.0)
            assert ds_lcc.canopy_intercepted_load.max() > 0.0
            assert ds_lcc.canopy_intercepted_snowfall.max() > 0.0
            assert ds_lcc.canopy_sublimation.max() > 0.0
            assert ds_lcc.canopy_melt.max() > 0.0
        else:
            assert np.all(np.isnan(ds_lcc.canopy_intercepted_load))
            assert np.all(np.isnan(ds_lcc.canopy_intercepted_snowfall))
            assert np.all(np.isnan(ds_lcc.canopy_sublimation))
            assert np.all(np.isnan(ds_lcc.canopy_melt))


@pytest.mark.slow
def test_canopy_meteorology(canopy_run):
    model = canopy_run
    ds = model.point_output.data
    lccs = model.config.land_cover.classes.keys()

    for lcc_num, lcc in enumerate(lccs):
        lcc_params = model.config.land_cover.classes[lcc]
        is_forest = lcc_params.get("is_forest", False)

        if is_forest:
            ds_lcc = ds.isel(point=lcc_num)
            assert np.all(ds_lcc.rel_hum >= 0.0)
            assert np.all(ds_lcc.rel_hum <= 100.0)
            assert np.all(ds_lcc.wind_speed >= 0.0)
            assert np.all(ds_lcc.sw_in >= 0.0)
            assert np.all(ds_lcc.lw_in >= 0.0)
            assert np.all(np.abs(ds_lcc.temp - ds_lcc.top_canopy_temp) <= 10.0)


def test_no_forest():
    config = base_config()
    config.start_date = "2019-10-01"
    config.end_date = "2019-10-01"

    model = oa.OpenAmundsen(config)
    model.initialize()

    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_path = Path(temp_dir)

        for p in Path(config.input_data.grids.dir).glob("*.asc"):
            (tmp_path / p.name).symlink_to(p)

        config.input_data.grids.dir = str(temp_dir)

        lc = np.zeros(model.grid.shape, dtype=int)
        rio_meta = {"driver": "AAIGrid"}
        oa.fileio.write_raster_file(
            oa.util.raster_filename("lc", config),
            lc.astype(np.int32),
            model.grid.transform,
            **rio_meta,
        )

        config.canopy.enabled = True
        model = oa.OpenAmundsen(config)
        model.initialize()
        model.run()


@pytest.mark.slow
@pytest.mark.comparison
def test_compare_canopy(canopy_run):
    compare_datasets(
        "canopy_point",
        canopy_run.point_output.data,
        point="point5",  # coniferous forest
    )



from openamundsen import constants


def initialize(model):
    """
    Initialize state variables for the soil model.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Cox, P. M., Betts, R. A., Bunton, C. B., Essery, R. L. H., Rowntree,
       P. R., & Smith, J. (1999). The impact of new land surface physics on the
       GCM simulation of climate and climate sensitivity. Climate Dynamics, 15(3),
       183–203.  https://doi.org/10.1007/s003820050276

    .. [2] Best, M. J., Pryor, M., Clark, D. B., Rooney, G. G., Essery, R. L. H.,
       Ménard, C. B., Edwards, J. M., Hendry, M. A., Porson, A., Gedney, N.,
       Mercado, L. M., Sitch, S., Blyth, E., Boucher, O., Cox, P. M., Grimmond, C.
       S. B., & Harding, R. J. (2011). The Joint UK Land Environment Simulator
       (JULES), model description – Part 1: Energy and water fluxes. Geoscientific
       Model Development, 4(3), 677–699. https://doi.org/10.5194/gmd-4-677-2011

    .. [3] Chadburn, S., Burke, E., Essery, R., Boike, J., Langer, M., Heikenfeld, M.,
       Cox, P., & Friedlingstein, P. (2015). An improved representation of
       physical permafrost dynamics in the JULES land-surface model. Geoscientific
       Model Development, 8(5), 1493–1508. https://doi.org/10.5194/gmd-8-1493-2015
    """
    roi = model.grid.roi
    cfg = model.config.soil
    s = model.state.soil

    # Temperature
    s.temp[:, roi] = cfg.init_temp

    # Thickness
    for i, thickness in enumerate(cfg.thickness):
        s.thickness[i, roi] = thickness

    # Volumetric heat capacity of dry soil
    s.vol_heat_cap_dry[roi] = (
        constants.VOL_HEAT_CAP_SAND * cfg.sand_fraction
        + constants.VOL_HEAT_CAP_CLAY * cfg.clay_fraction
    ) / (cfg.sand_fraction + cfg.clay_fraction)

    # Clapp-Hornberger exponent b
    s.clapp_hornberger[roi] = 3.1 + 15.7 * cfg.clay_fraction - 0.3 * cfg.sand_fraction

    # Saturated soil water pressure
    s.sat_water_pressure[roi] = 10 ** (0.17 - 0.63 * cfg.clay_fraction - 1.58 * cfg.sand_fraction)

    # Volumetric soil moisture content at saturation
    # (values from doi:10.1029/wr020i006p00682, table 4)
    s.vol_moisture_content_sat[roi] = 0.505 - 0.037 * cfg.clay_fraction - 0.142 * cfg.sand_fraction

    # Volumetric soil moisture content at critical point
    s.vol_moisture_content_crit[roi] = s.vol_moisture_content_sat[roi] * (
        s.sat_water_pressure[roi] / 3.364  # soil suction of 3.364 m
    ) ** (1.0 / s.clapp_hornberger[roi])

    # Thermal conductivity of soil minerals
    s.therm_cond_minerals[roi] = (
        (constants.THERM_COND_CLAY**cfg.clay_fraction)
        * (
            constants.THERM_COND_SAND ** (1 - cfg.clay_fraction)
        )  # TODO really 1 - clay_fraction and not just sand_fraction?
    )

    # Thermal conductivity of dry soil
    s.therm_cond_dry[roi] = (
        constants.THERM_COND_AIR ** s.vol_moisture_content_sat[roi]
    ) * s.therm_cond_minerals[roi] ** (1 - s.vol_moisture_content_sat[roi])

    # Volumetric soil moisture content
    s.vol_moisture_content[:, roi] = cfg.init_moisture_content * s.vol_moisture_content_sat[roi]



from __future__ import annotations

from typing import Literal

import numpy as np
import pandas as pd
import pwlf
from loguru import logger

from openamundsen import constants, interpolation, meteo, util


def _param_station_data(ds, param, date):
    """
    Return station measurements including station x, y and z coordinates for one or more given
    parameter(s) and a specified date.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset containing the station measurements (i.e. the model.meteo
        variable)

    param : str or list
        Parameter(s) for which the measurements should be returned.

    date : datetime-like
        Date for which the measurements should be returned.

    Returns
    -------
    data : ndarray
        Measurements (excluding nodata values). If param is a string, this is a 1-D array; if param
        is a list of strings it is a 2-D array with dimensions
        (len(param), <number of available measurements where all parameters are non-null>).
    x, y, z : ndarrays
        Corresponding x, y and z coordinates.
    """
    if isinstance(param, str):
        param_as_list = [param]
    else:
        param_as_list = param

    ds_param = ds[[*param_as_list, "x", "y", "alt"]].sel(time=date).dropna(dim="station")
    data = ds_param[param]
    xs = ds_param["x"].values
    ys = ds_param["y"].values
    zs = ds_param["alt"].values

    if isinstance(param, str):
        data_vals = data.values
    else:
        data_vals = data.to_array().values

    return data_vals, xs, ys, zs


def _linear_fit(x, y):
    """
    Calculate the linear regression for a set of sample points.

    Parameters
    ----------
    x, y : ndarray
        x and y coordinates of the sample points.

    Returns
    -------
    slope, intercept : float
    """
    if len(x) < 2:
        slope = 0
        intercept = 0
    else:
        slope, intercept = np.polyfit(x, y, 1)

    return slope, intercept


def _piecewise_linear_fit(
    xs: np.ndarray,
    ys: np.ndarray,
    max_segments: int,
    min_points_per_segment: int,
    delta_aic_threshold: float,
) -> dict | None:
    best = None

    for num_segments in range(1, max_segments + 1):
        # Stop if it is impossible to allocate min_points_per_segment to each segment
        if num_segments > 1 and len(xs) < num_segments * min_points_per_segment:
            break

        pw = pwlf.PiecewiseLinFit(xs, ys)
        try:
            pw.fit(
                num_segments,
                seed=42,
                maxiter=50,
                popsize=15,
            )
        except (ValueError, ZeroDivisionError):
            continue

        breaks = pw.fit_breaks

        # Ensure there are at least min_points_per_segment points in each segment (except for the
        # single-segment case)
        if num_segments > 1 and any(
            np.sum((xs >= lo) & (xs <= hi)) < min_points_per_segment
            for lo, hi in zip(breaks[:-1], breaks[1:])
        ):
            break

        preds = pw.predict(xs)
        rss = np.sum((ys - preds) ** 2)
        n = len(ys)
        k = 2 * num_segments
        aic = n * np.log(rss / n) + 2 * k

        model_params = {
            "model": pw,
            "breaks": breaks,
            "aic": aic,
        }

        if best is None:
            best = model_params
            continue

        model_has_improved = (best["aic"] - aic) >= delta_aic_threshold
        if model_has_improved:
            best = model_params
        else:
            # If the model has not improved compared to the earlier one we stop, since it is
            # unlikely in this case that the next one will improve the results
            break

    return best


def _apply_linear_trend(data, elevs, trend, direction):
    """
    Detrend or retrend a set of data points using a given trend value.

    Parameters
    ----------
    data : ndarray
        Detrended values.

    elevs : ndarray
        Elevations corresponding to the data points.

    trend : float
        Trend value.

    direction : str
        Either "detrend" or "retrend".

    Returns
    -------
    data_trend : ndarray
        De-/retrended data points.
    """
    if direction == "detrend":
        return data - trend * elevs
    elif direction == "retrend":
        return data + trend * elevs
    else:
        raise NotImplementedError(f"Unsupported direction: {direction}")


def _apply_piecewise_trend(
    data: np.ndarray,
    elevs: np.ndarray,
    pw_model: pwlf.PiecewiseLinFit,
    direction: Literal["detrend", "retrend"],
) -> np.ndarray:
    """
    Add or remove a piecewise-linear elevation trend.

    Parameters
    ----------
    data
        Values to (de)trend.
    elevs
        Matching elevations.
    pw_model
        Fitted pwlf.PiecewiseLinFit.
    direction
        Either "detrend" or "retrend".

    Returns
    -------
    np.ndarray
        De-/retrended data points.
    """
    trend_vals = pw_model.predict(elevs)

    if direction == "detrend":
        return data - trend_vals
    elif direction == "retrend":
        return data + trend_vals
    else:
        raise ValueError(f"Unsupported direction: {direction}")


def interpolate_station_data(model):
    """
    Interpolate station measurements to the model grid.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.
    """
    logger.debug("Interpolating station data")

    date = model.date
    roi = model.grid.roi
    target_xs = model.grid.roi_points[:, 0]
    target_ys = model.grid.roi_points[:, 1]
    target_zs = model.state.base.dem[roi]

    model._interpolation = {}

    for param in ("temp", "precip"):
        param_config = model.config["meteo"]["interpolation"][
            constants.INTERPOLATION_CONFIG_PARAM_MAPPINGS[param]
        ]
        data, xs, ys, zs = _param_station_data(model.meteo, param, date)

        model.state.meteo[param][roi], model._interpolation[param] = interpolate_param(
            param,
            model.date,
            param_config,
            data,
            xs,
            ys,
            zs,
            target_xs,
            target_ys,
            target_zs,
        )

    # For relative humidity dew point temperature is interpolated, so we also need station
    # temperatures and the interpolated temperature field
    param = "rel_hum"
    param_config = model.config["meteo"]["interpolation"][
        constants.INTERPOLATION_CONFIG_PARAM_MAPPINGS[param]
    ]
    data, xs, ys, zs = _param_station_data(model.meteo, [param, "temp"], date)
    rel_hums = data[0, :]
    temps = data[1, :]
    model.state.meteo[param][roi], _ = interpolate_param(
        param,
        model.date,
        param_config,
        rel_hums,
        xs,
        ys,
        zs,
        target_xs,
        target_ys,
        target_zs,
        temps=temps,
        target_temps=model.state.meteo["temp"][roi],
    )

    wind_config = model.config["meteo"]["interpolation"]["wind"]
    if wind_config["method"] == "idw":
        data, xs, ys, zs = _param_station_data(model.meteo, "wind_speed", date)
        model.state.meteo["wind_speed"][roi], _ = interpolate_param(
            "wind_speed",
            model.date,
            wind_config,
            data,
            xs,
            ys,
            zs,
            target_xs,
            target_ys,
            target_zs,
        )

        if model._has_wind_gusts:
            data, xs, ys, zs = _param_station_data(model.meteo, "wind_speed_gust", date)
            model.state.meteo["wind_speed_gust"][roi], _ = interpolate_param(
                "wind_speed",
                model.date,
                wind_config,
                data,
                xs,
                ys,
                zs,
                target_xs,
                target_ys,
                target_zs,
            )
    elif wind_config["method"] == "liston":
        data, xs, ys, zs = _param_station_data(model.meteo, ["wind_speed", "wind_dir"], date)
        wind_speeds = data[0, :]
        wind_dirs = data[1, :]
        wind_speed_corr, wind_dir_corr = _liston_wind_correction(
            model.date,
            wind_config,
            wind_speeds,
            wind_dirs,
            xs,
            ys,
            zs,
            target_xs,
            target_ys,
            target_zs,
            model.state.base.slope[roi],
            model.state.base.aspect[roi],
            model.state.base.scaled_curvature[roi],
        )
        model.state.meteo.wind_speed[roi] = np.clip(
            wind_speed_corr,
            *constants.ALLOWED_METEO_VAR_RANGES["wind_speed"],
        )
        model.state.meteo.wind_dir[roi] = wind_dir_corr

        if model._has_wind_gusts:
            data, xs, ys, zs = _param_station_data(
                model.meteo,
                ["wind_speed_gust", "wind_dir"],
                date,
            )
            wind_speed_gusts = data[0, :]
            wind_dirs = data[1, :]
            wind_speed_gusts_corr, _ = _liston_wind_correction(
                model.date,
                wind_config,
                wind_speed_gusts,
                wind_dirs,
                xs,
                ys,
                zs,
                target_xs,
                target_ys,
                target_zs,
                model.state.base.slope[roi],
                model.state.base.aspect[roi],
                model.state.base.scaled_curvature[roi],
            )
            model.state.meteo.wind_speed_gust[roi] = np.clip(
                wind_speed_gusts_corr,
                *constants.ALLOWED_METEO_VAR_RANGES["wind_speed_gust"],
            )
    else:
        raise NotImplementedError(f"Unsupported method: {wind_config.method}")


def interpolate_param(
    param: str,
    date: pd.Timestamp,
    param_config: dict,
    data: np.ndarray,
    xs: np.ndarray,
    ys: np.ndarray,
    zs: np.ndarray,
    target_xs: np.ndarray,
    target_ys: np.ndarray,
    target_zs: np.ndarray,
    temps: np.ndarray | None = None,
    target_temps: np.ndarray | None = None,
) -> tuple[np.ndarray, dict]:
    """
    Interpolate a set of data points to a set of target points.

    Parameters
    ----------
    param : str
        Parameter to be interpolated. Can be one of 'temp', 'precip', 'rel_hum', 'wind_speed',
        'wind_vec'.

    date : datetime-like
        Date (required in the case of seasonally specified lapse rates).

    param_config : dict
        Interpolation configuration for the respective parameter from the model run configuration
        (e.g., config['meteo']['interpolation']['temperature']).

    data : ndarray
        Values to be interpolated.

    xs, ys, zs : ndarray
        x, y and z coordinates of the stations.

    target_xs, target_ys, target_zs : ndarray
        x, y and z coordinates of the interpolation targets.

    temps : ndarray, default None
        Temperatures at the source locations (required for humidity interpolation).

    target_temps : ndarray, default None
        Temperatures at the target locations (required for humidity interpolation).

    Returns
    -------
    data_interpol : ndarray
        Interpolated values for the target locations.
    """
    if param not in ("temp", "precip", "rel_hum", "wind_speed", "wind_vec"):
        raise ValueError(f"Unsupported parameter: {param}")

    trend_method = param_config["trend_method"]
    lapse_rate = param_config["lapse_rate"][date.month - 1]

    interpolation_params = {"gradient_type": "none"}

    # If there are no points to be interpolated, return an all-zero array for precipitation, and an
    # all-nan array for all other parameters
    if data.size == 0:
        if param == "precip":
            fill_value = 0.0
        else:
            fill_value = np.nan

        return (np.full(target_xs.shape, fill_value), interpolation_params)

    # For relative humidity interpolate dew point temperature and convert back to humidity later
    if param == "rel_hum":
        if temps is None or target_temps is None:
            raise ValueError("Temperature must be provided for humidity interpolation")

        rel_hums = data
        dew_point_temps = meteo.dew_point_temperature(temps, rel_hums)
        data = dew_point_temps

    regression_params = param_config.get("regression_params", {})
    max_segments = regression_params.get("max_segments", 1)

    if trend_method in ("regression", "fixed", "fractional") and max_segments == 1:
        # Use simple linear regression
        if trend_method == "regression":
            # When using the regression method, the passed lapse rate is overwritten
            lapse_rate, _ = _linear_fit(zs, data)
        elif trend_method == "fixed":
            pass  # do nothing, i.e., use the passed lapse rate as is
        elif trend_method == "fractional":
            lapse_rate *= np.nanmean(data)

        data_detrended = _apply_linear_trend(data, zs, lapse_rate, "detrend")
        data_detrended_interpol = interpolation.idw(
            xs,
            ys,
            data_detrended,
            target_xs,
            target_ys,
        )
        data_interpol = _apply_linear_trend(
            data_detrended_interpol,
            target_zs,
            lapse_rate,
            "retrend",
        )
        interpolation_params.update(
            gradient_type="linear",
            gradient=lapse_rate,
        )
    elif trend_method == "regression" and max_segments > 1:
        # Use piecewise linear regression
        pw_lr = _piecewise_linear_fit(
            zs,
            data,
            regression_params["max_segments"],
            regression_params["min_points_per_segment"],
            regression_params["delta_aic_threshold"],
        )

        if pw_lr is None:
            return (np.full(target_xs.shape, np.nan), interpolation_params)

        pw_model = pw_lr["model"]
        data_detrended = _apply_piecewise_trend(data, zs, pw_model, "detrend")
        data_detrended_interpol = interpolation.idw(
            xs,
            ys,
            data_detrended,
            target_xs,
            target_ys,
        )
        data_interpol = _apply_piecewise_trend(
            data_detrended_interpol,
            target_zs,
            pw_model,
            "retrend",
        )

        interpolation_params["gradient_type"] = "piecewise_linear"
        if pw_lr is not None:
            interpolation_params.update(**pw_lr)
    elif trend_method == "adjustment_factor":
        data_interpol_notrend = interpolation.idw(xs, ys, data, target_xs, target_ys)
        zs_interpol = interpolation.idw(xs, ys, zs, target_xs, target_ys)
        z_diffs = target_zs - zs_interpol
        data_interpol = data_interpol_notrend * (  # eq. (33) from [1]
            (1 + lapse_rate * z_diffs) / (1 - lapse_rate * z_diffs)
        )
        interpolation_params.update(
            gradient_type="adjustment_factor",
            gradient=lapse_rate,
        )
    else:
        raise ValueError

    if param == "rel_hum":
        target_dew_point_temps = data_interpol
        vapor_press = meteo.saturation_vapor_pressure(target_dew_point_temps, "water")
        sat_vapor_press = meteo.saturation_vapor_pressure(target_temps, "water")
        data_interpol = 100 * vapor_press / sat_vapor_press

    # Restrict interpolated values to the range of the point values if extrapolation is disabled in
    # the parameter config
    if not param_config["extrapolate"] and len(data) > 0:
        min_range = np.nanmin(data)
        max_range = np.nanmax(data)

        if np.isfinite(min_range):
            data_interpol = np.clip(data_interpol, min_range, max_range)

    if param in constants.ALLOWED_METEO_VAR_RANGES:
        min_range, max_range = constants.ALLOWED_METEO_VAR_RANGES[param]
        data_interpol = np.clip(data_interpol, min_range, max_range)

    return data_interpol, interpolation_params


def _liston_wind_correction(
    date,
    wind_config,
    wind_speeds,
    wind_dirs,
    xs,
    ys,
    zs,
    target_xs,
    target_ys,
    target_zs,
    target_slopes,
    target_aspects,
    target_scaled_curvatures,
):
    wind_us, wind_vs = meteo.wind_to_uv(wind_speeds, wind_dirs)
    wind_u_roi, _ = interpolate_param(
        "wind_vec",
        date,
        wind_config,
        wind_us,
        xs,
        ys,
        zs,
        target_xs,
        target_ys,
        target_zs,
    )
    wind_v_roi, _ = interpolate_param(
        "wind_vec",
        date,
        wind_config,
        wind_vs,
        xs,
        ys,
        zs,
        target_xs,
        target_ys,
        target_zs,
    )
    wind_speed_roi, wind_dir_roi = meteo.wind_from_uv(wind_u_roi, wind_v_roi)
    wind_dir_minus_aspect = np.deg2rad(wind_dir_roi - target_aspects)
    wind_slope = (  # slope in the direction of the wind (eq. (15))
        np.deg2rad(target_slopes) * np.cos(wind_dir_minus_aspect)
    )
    wind_slope_scaled = util.normalize_array(wind_slope, -0.5, 0.5)
    wind_weighting_factor = (  # eq. (16)
        1
        + wind_config.slope_weight * wind_slope_scaled
        + wind_config.curvature_weight * target_scaled_curvatures
    )
    wind_dir_diverting_factor = (  # eq. (18)
        -0.5 * wind_slope_scaled * np.sin(-2 * wind_dir_minus_aspect)
    )
    wind_speed_corr = wind_speed_roi * wind_weighting_factor
    wind_dir_corr = (wind_dir_roi + wind_dir_diverting_factor) % 360

    return wind_speed_corr, wind_dir_corr



from .griddedoutput import GriddedOutputManager
from .meteo import (
    read_csv_meteo_file,
    read_meteo_data,
    read_netcdf_meteo_file,
)
from .pointoutput import PointOutputManager
from .raster import (
    read_raster_file,
    read_raster_metadata,
    write_raster_file,
)

__all__ = [
    "GriddedOutputManager",
    "PointOutputManager",
    "read_csv_meteo_file",
    "read_meteo_data",
    "read_netcdf_meteo_file",
    "read_raster_file",
    "read_raster_metadata",
    "write_raster_file",
]



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from numpy.testing import assert_array_equal

import openamundsen as oa

from .conftest import base_config


def compare_tsp(tsp, **kwargs):
    d = {
        "first_of_run": False,
        "strict_first_of_year": False,
        "strict_first_of_month": False,
        "strict_first_of_day": False,
        "first_of_year": False,
        "first_of_month": False,
        "first_of_day": False,
        "last_of_run": False,
        "strict_last_of_year": False,
        "strict_last_of_month": False,
        "strict_last_of_day": False,
        "last_of_year": False,
        "last_of_month": False,
        "last_of_day": False,
    }
    d.update(kwargs)

    for key in d:
        assert getattr(tsp, key) == d[key], f"{key} should be {d[key]}"


def test_timestep_properties():
    config = base_config()
    config.start_date = "2015-07-28 00:00"
    config.end_date = "2015-12-31"
    config.timestep = "h"

    model = oa.OpenAmundsen(config)
    model.initialize()

    model.run_single()
    compare_tsp(
        model.timestep_props,
        first_of_run=True,
        first_of_year=True,
        first_of_month=True,
        first_of_day=True,
        strict_first_of_day=True,
    )

    model.run_single()
    compare_tsp(model.timestep_props)

    while model.date < pd.Timestamp("2015-07-29 00:00"):
        model.run_single()
    compare_tsp(
        model.timestep_props,
        first_of_day=True,
        strict_first_of_day=True,
    )

    while model.date < pd.Timestamp("2015-08-01 00:00"):
        model.run_single()
    compare_tsp(
        model.timestep_props,
        first_of_month=True,
        first_of_day=True,
        strict_first_of_month=True,
        strict_first_of_day=True,
    )

    config.start_date = "2015-07-28 23:00"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run_single()
    compare_tsp(
        model.timestep_props,
        first_of_run=True,
        first_of_year=True,
        first_of_month=True,
        first_of_day=True,
        last_of_day=True,
        strict_last_of_day=True,
    )

    config.start_date = "2015-08-01 00:00"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run_single()
    compare_tsp(
        model.timestep_props,
        first_of_run=True,
        first_of_year=True,
        first_of_month=True,
        first_of_day=True,
        strict_first_of_month=True,
        strict_first_of_day=True,
    )

    config.start_date = "2014-12-31 23:00"
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run_single()
    compare_tsp(
        model.timestep_props,
        first_of_run=True,
        first_of_year=True,
        first_of_month=True,
        first_of_day=True,
        last_of_year=True,
        last_of_month=True,
        last_of_day=True,
        strict_last_of_year=True,
        strict_last_of_month=True,
        strict_last_of_day=True,
    )

    model.run_single()
    compare_tsp(
        model.timestep_props,
        first_of_year=True,
        first_of_month=True,
        first_of_day=True,
        strict_first_of_year=True,
        strict_first_of_month=True,
        strict_first_of_day=True,
    )

    config.end_date = "2015-12-15 12:00"
    config.start_date = config.end_date
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run_single()
    compare_tsp(
        model.timestep_props,
        first_of_run=True,
        last_of_run=True,
        first_of_year=True,
        first_of_month=True,
        first_of_day=True,
        last_of_year=True,
        last_of_month=True,
        last_of_day=True,
    )


@pytest.mark.slow
def test_state_variable_reset():
    config = base_config()
    config.start_date = "2019-10-01"
    config.end_date = "2020-05-31"
    config.timestep = "3h"
    config.evapotranspiration.enabled = True
    config.canopy.enabled = True

    model = oa.OpenAmundsen(config)
    model.initialize()

    for category in model.state.categories:
        for var_name in model.state[category]._meta.keys():
            full_var_name = f"{category}.{var_name}"
            config.output_data.grids.variables.append(
                {
                    "var": full_var_name,
                    "name": full_var_name,
                }
            )

    config.reset_state_variables = False
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds0 = model.gridded_output.data

    config.reset_state_variables = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    ds1 = model.gridded_output.data

    compare_vars = [v["name"] for v in config.output_data.grids.variables]
    success = True

    for v in compare_vars:
        arr0 = ds0[v].values
        arr1 = ds1[v].values

        if not np.allclose(arr0, arr1, equal_nan=True):
            print(f"Mismatch in variable {v}\nWithout reset:\n{arr0}\nWith reset:\n{arr1}")
            success = False

    assert success


def test_simulation_timezone(tmp_path):
    ds = xr.load_dataset(f"{pytest.DATA_DIR}/meteo/rofental/netcdf/proviantdepot.nc")
    ds.to_netcdf(tmp_path / "proviantdepot.nc")

    config = base_config()
    config.timestep = "h"
    config.input_data.meteo.dir = str(tmp_path)

    config.start_date = "2020-01-15 00:00"
    config.end_date = "2020-01-15 23:00"
    config.simulation_timezone = None
    model1 = oa.OpenAmundsen(config)
    model1.initialize()
    model1.run()

    ds.shift(time=-1).to_netcdf(tmp_path / "proviantdepot.nc")

    config.start_date = "2020-01-14 23:00"
    config.end_date = "2020-01-15 22:00"
    config.simulation_timezone = 0
    model2 = oa.OpenAmundsen(config)
    model2.initialize()
    model2.run()

    ds1 = model1.point_output.data
    ds2 = model2.point_output.data
    ds2["time"] = ds2["time"] + pd.Timedelta(hours=1)
    xr.testing.assert_identical(ds1, ds2)


def test_roi(tmp_path):
    config = base_config()
    config.start_date = "2020-01-15 12:00"
    config.end_date = "2020-01-15 12:00"

    for p in Path(config.input_data.grids.dir).glob("*.asc"):
        if not p.name.startswith("roi_"):
            (tmp_path / p.name).symlink_to(p)

    config.input_data.grids.dir = str(tmp_path)

    # No ROI set
    model1 = oa.OpenAmundsen(config)
    model1.initialize()
    assert np.all(model1.grid.roi)
    model1.run()

    # ROI manually set
    roi = model1.grid.roi.copy()
    roi[:10, :10] = False
    oa.fileio.write_raster_file(
        oa.util.raster_filename("roi", config),
        roi.astype(np.uint8),
        model1.grid.transform,
        driver="AAIGrid",
        dtype="uint8",
    )
    model2 = oa.OpenAmundsen(config)
    model2.initialize()
    assert_array_equal(roi, model2.grid.roi)
    model2.run()

    # ROI should be False where DEM is NaN
    Path(oa.util.raster_filename("roi", config)).unlink()
    dem = model2.state.base.dem
    dem[-10:, -10:] = np.nan
    oa.fileio.write_raster_file(
        oa.util.raster_filename("dem", config),
        dem,
        model2.grid.transform,
        driver="AAIGrid",
    )
    model3 = oa.OpenAmundsen(config)
    model3.initialize()
    assert np.all(model3.grid.roi[np.isfinite(dem)])
    assert not np.any(model3.grid.roi[np.isnan(dem)])
    model3.run()

    for model in (model1, model2, model3):
        for var in ("meteo.temp", "snow.swe"):
            var_data = model.state[var]
            assert np.all(np.isfinite(var_data[model.grid.roi]))
            assert np.all(np.isnan(var_data[~model.grid.roi]))


def test_extend_roi_with_stations(tmp_path):
    config = base_config()
    config.start_date = "2020-07-15 12:00"
    config.end_date = "2020-07-15 12:00"

    model = oa.OpenAmundsen(config)
    model.initialize()

    for p in Path(config.input_data.grids.dir).glob("*.asc"):
        if not p.name.startswith("roi_"):
            (tmp_path / p.name).symlink_to(p)

    config.input_data.grids.dir = str(tmp_path)

    station_id = "proviantdepot"
    meteo_station = model.meteo.sel(station=station_id)
    row = meteo_station.row
    col = meteo_station.col

    roi = model.grid.roi
    roi[row, col] = False
    oa.fileio.write_raster_file(
        oa.util.raster_filename("roi", config),
        roi.astype(np.uint8),
        model.grid.transform,
        driver="AAIGrid",
        dtype="uint8",
    )

    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    assert not model.meteo.within_roi.loc[station_id]
    assert np.isnan(model.state.meteo.temp[row, col])
    assert np.isnan(model.state.meteo.sw_in[row, col])

    config.extend_roi_with_stations = True
    model = oa.OpenAmundsen(config)
    model.initialize()
    model.run()
    assert model.meteo.within_roi.loc[station_id]
    assert np.isfinite(model.state.meteo.temp[row, col])
    assert np.isfinite(model.state.meteo.sw_in[row, col])



import numpy as np
from numpy.testing import assert_allclose, assert_array_equal

import openamundsen.meteo as meteo


def test_atmospheric_pressure():
    elevs = [0, 1000, 2000, 3000]
    pressures = [101_325, 89_874.57, 79_495.22, 70_108.54]  # values from https://www.mide.com/air-pressure-at-altitude-calculator # fmt: skip # noqa: E501
    press_calc = meteo.atmospheric_pressure(elevs)
    assert_allclose(press_calc, pressures, rtol=1e-2)


def test_pressure_to_altitude():
    elevs = np.linspace(0, 3000, 100)
    pressures = meteo.atmospheric_pressure(elevs)
    elevs_calc = meteo.pressure_to_altitude(pressures)
    assert_allclose(elevs, elevs_calc)


def test_latent_heat_of_vaporization():
    assert_allclose(
        meteo.latent_heat_of_vaporization([273.15, 20 + 273.15]),
        [2.501e6, 2.45e6],
        rtol=2e-2,
    )


def test_saturation_vapor_pressure():
    temps = np.array([-40, -20, 0, 20, 40]) + 273.15
    assert_allclose(
        meteo.saturation_vapor_pressure(temps),
        [12.84, 103.24, 611.15, 2339.32, 7384.94],
        rtol=1e-2,
    )


def test_absolute_humidity():
    temps = np.array([30, 20, 10, 0]) + 273.15
    rel_hums = np.array([10, 30, 70, 20])
    abs_hums = np.array([3.0, 5.2, 6.6, 1.0]) * 1e-3  # values from https://www.tis-gdv.de/tis_e/misc/klima-htm/ # fmt: skip # noqa: E501
    vapor_pressures = meteo.vapor_pressure(temps, rel_hums)
    assert_allclose(
        meteo.absolute_humidity(temps, vapor_pressures),
        abs_hums,
        rtol=5e-2,
    )


def test_psychrometric_constant():
    temps = np.array([0, 5, 10, 20]) + 273.15
    psych_consts = np.array([65.5, 65.8, 66.1, 66.8])  # from http://ponce.sdsu.edu/psychrometric_constant.html # fmt: skip # noqa: E501

    lat_heat_vaps = meteo.latent_heat_of_vaporization(temps)
    atmos_pressures = meteo.atmospheric_pressure(0)
    vapor_pressures = meteo.saturation_vapor_pressure(temps)
    spec_hums = meteo.specific_humidity(atmos_pressures, vapor_pressures)
    spec_heat_caps = meteo.specific_heat_capacity_moist_air(spec_hums)

    assert_allclose(
        meteo.psychrometric_constant(atmos_pressures, spec_heat_caps, lat_heat_vaps),
        psych_consts,
        atol=0.75,
    )


def test_wet_bulb_temperature():
    temps = np.array([5, 10, 20, 30]) + 273.15
    rel_hums = np.array([50, 70, 90, 100])
    wet_bulb_temps = (np.array([1.345, 7.393, 18.863, 30]) + 273.15)  # from http://www.flycarpet.net/en/PsyOnline # fmt: skip # noqa: E501
    # TODO also test negative temps

    # WBT should be nan where either T or RH are nan
    temps = np.append(temps, [np.nan, 273.15])
    rel_hums = np.append(rel_hums, [100, np.nan])
    wet_bulb_temps = np.append(wet_bulb_temps, [np.nan, np.nan])

    lat_heat_vaps = meteo.latent_heat_of_vaporization(temps)
    atmos_pressures = meteo.atmospheric_pressure(0)
    vapor_pressures = meteo.vapor_pressure(temps, rel_hums)
    spec_hums = meteo.specific_humidity(atmos_pressures, vapor_pressures)
    spec_heat_caps = meteo.specific_heat_capacity_moist_air(spec_hums)
    psych_consts = meteo.psychrometric_constant(atmos_pressures, spec_heat_caps, lat_heat_vaps)

    assert_allclose(
        meteo.wet_bulb_temperature(temps, rel_hums, vapor_pressures, psych_consts),
        wet_bulb_temps,
        atol=0.1,
    )


def test_relative_humidity():
    temps = np.array([30, 20, 10, 0]) + 273.15
    rel_hums = np.array([10, 30, 70, 20])
    vapor_pressures = meteo.vapor_pressure(temps, rel_hums)
    abs_hums = meteo.absolute_humidity(temps, vapor_pressures)

    assert_allclose(
        rel_hums,
        meteo.relative_humidity(temps, abs_hums),
        atol=1e-6,
    )


def test_dew_point_temperature():
    temps = np.array([-5, 5, 15, 20, 25, 30, 35]) + 273.15
    rel_hums = np.array([30, 50, 50, 70, 90, 100, 30])
    tds = np.array([-19.8, -4.6, 4.65, 14.36, 23.24, 30, 14.84]) + 273.15

    assert_allclose(
        meteo.dew_point_temperature(temps, rel_hums),
        tds,
        atol=0.1,
    )


def test_cloud_factor_from_cloud_fraction():
    cloud_fractions = np.linspace(0, 1, 100)
    cloud_factors = meteo.cloud_factor_from_cloud_fraction(cloud_fractions)
    assert cloud_factors.min() >= 0
    assert cloud_factors.max() <= 1


def test_cloud_fraction_from_cloud_factor():
    cloud_fractions = np.linspace(0, 1, 100)
    cloud_factors = meteo.cloud_factor_from_cloud_fraction(cloud_fractions)
    cloud_fractions_inv = meteo.cloud_fraction_from_cloud_factor(cloud_factors)
    assert_allclose(cloud_fractions, cloud_fractions_inv, atol=1e-6)


def test_cloud_fraction_from_humidity():
    temps = [285, 277, 260.3, 268]
    rel_hums = [70, 0.1, 100, 33.77]
    elevs = [100, 500, 3000, 1400]

    cf = meteo.cloud_fraction_from_humidity(
        temps,
        rel_hums,
        elevs,
        -0.0026,
        -0.0044,
    )

    assert cf.min() >= 0
    assert cf.max() <= 1
    assert_allclose(cf[[1, 2]], [0.075224, 0.845268], atol=1e-3)


def test_precipitation_phase():
    temps = np.array([-2, 4, -1.2, 0, 23.4, -24.5, -0.01, 0.01]) + 273.15
    sf = meteo.precipitation_phase(temps, threshold_temp=273.15)
    assert_array_equal(sf, [1, 0, 1, 0, 0, 1, 1, 0])

    temps = np.array([-0.99, -1.01, -1.0, 0.0, 1.0, 2.99, 3.01, 3.0]) + 273.15
    sf = meteo.precipitation_phase(temps, threshold_temp=1 + 273.15, temp_range=4)
    assert_allclose(sf, [0.9975, 1, 1, 0.75, 0.5, 0.0025, 0, 0], atol=1e-10)


def test_log_wind_profile():
    assert_allclose(meteo.log_wind_profile(1, 10, 2, 0.03), 0.722947444)


def test_wind_uv():
    u = np.array([0, -1, 0, 1, 1, 1, -1, -1])
    v = np.array([-1, 0, 1, 0, 1, -1, 1, -1])
    ws, wd = meteo.wind_from_uv(u, v)
    assert_allclose(ws, [1, 1, 1, 1, np.sqrt(2), np.sqrt(2), np.sqrt(2), np.sqrt(2)])
    assert_allclose(wd, [0, 90, 180, 270, 225, 315, 135, 45])
    u2, v2 = meteo.wind_to_uv(ws, wd)
    assert_allclose(u, u2, atol=1e-10)
    assert_allclose(v, v2, atol=1e-10)



import os
import subprocess
import textwrap

import pytest
import xarray as xr

from openamundsen.util import to_yaml

from .conftest import base_config


def test_numba(tmp_path):
    config = base_config()
    config.start_date = "2020-01-18"
    config.end_date = "2020-01-18"
    config.results_dir = tmp_path
    config.output_data.timeseries.format = "netcdf"
    config.output_data.grids.format = "netcdf"
    config.output_data.grids.variables = [
        {"var": "meteo.temp"},
        {"var": "snow.swe"},
        {"var": "snow.temp", "name": "snow_temp"},
    ]

    config_file = tmp_path / "config.yml"
    with open(config_file, "w") as f:
        f.write(to_yaml(config.toDict()))

    subprocess.check_call(["openamundsen", str(config_file)])  # noqa: S603, S607
    ds_points_numba = xr.load_dataset(tmp_path / "output_timeseries.nc")
    ds_grids_numba = xr.load_dataset(tmp_path / "output_grids.nc")

    env = os.environ.copy()
    env["NUMBA_DISABLE_JIT"] = "1"
    subprocess.check_call(["openamundsen", str(config_file)], env=env)  # noqa: S603, S607
    ds_points_nonumba = xr.load_dataset(tmp_path / "output_timeseries.nc")
    ds_grids_nonumba = xr.load_dataset(tmp_path / "output_grids.nc")

    assert ds_points_numba.equals(ds_points_nonumba)
    assert ds_grids_numba.equals(ds_grids_nonumba)


@pytest.mark.slow
def test_floating_errors(tmp_path):
    config = base_config()
    config.results_dir = tmp_path

    config_file = tmp_path / "config.yml"
    with open(config_file, "w") as f:
        f.write(to_yaml(config.toDict()))

    with open(tmp_path / "run.py", "w") as f:
        f.write(
            textwrap.dedent(
                f"""
            import numpy as np
            np.seterr(all='raise')

            import openamundsen as oa

            config = oa.read_config('{config_file}')
            model = oa.OpenAmundsen(config)
            model.initialize()
            model.run()
            """
            )
        )

    env = os.environ.copy()
    env["NUMBA_DISABLE_JIT"] = "1"
    subprocess.check_call(["python", tmp_path / "run.py"], env=env)  # noqa: S603, S607



import numpy as np
from loguru import logger

from openamundsen import constants as c
from openamundsen.modules.snow import CryoLayerID

MIN_ICE_THICKNESS = 10.0  # minimum required ice thickness for applying the delta h function (m)
TERMINUS_MASS_BALANCE_ELEVATION_PERCENTAGE = 10.0  # (%)
TERMINUS_ELEVATION_BUFFER = 50.0  # (m)


class GlacierModel:
    """
    Glacier geometry update model following [1].

    References
    ----------
    .. [1] Huss, M., Jouvet, G., Farinotti, D., & Bauder, A. (2010). Future
       high-mountain hydrology: A new parameterization of glacier retreat.
       Hydrology and Earth System Sciences, 14(5), 815–829.
       https://doi.org/10.5194/hess-14-815-2010
    """

    def __init__(self, model):
        self.model = model
        s = model.state
        sg = s.add_category("glaciers")
        sg.add_variable("glacier", long_name="Glacier ID", dtype=int)
        self.mass_balance_start_we = None

    def initialize(self):
        model = self.model
        roi = model.grid.roi
        s = self.model.state
        sg = s.glaciers
        ids = np.unique(sg.glacier[roi])
        self.glacier_ids = ids[ids > 0]

    def update(self):
        model = self.model
        s = model.state
        timestep_props = model.timestep_props
        mb_start_month = model.config.glaciers.delta_h.mass_balance_year_start_month
        mb_end_month = mb_start_month - 1 if mb_start_month > 1 else 12

        if model.date.month == mb_start_month and timestep_props.strict_first_of_month:
            self.mass_balance_start_we = s.snow.ice_content + s.snow.liquid_water_content
        elif (
            model.date.month == mb_end_month
            and timestep_props.strict_last_of_month
            and self.mass_balance_start_we is not None
        ):
            self._update_geometry()

    def _update_geometry(self):
        model = self.model
        s = model.state

        mass_balance_end_we = s.snow.ice_content + s.snow.liquid_water_content
        mb_layers = mass_balance_end_we - self.mass_balance_start_we
        ice_mb = mb_layers[CryoLayerID.ICE, :]
        firn_mb = mb_layers[CryoLayerID.FIRN, :]
        mb = ice_mb + firn_mb
        old_ice_thickness = self.mass_balance_start_we[CryoLayerID.ICE, :] / c.ICE_DENSITY
        new_ice_thickness = s.snow.thickness[CryoLayerID.ICE, :].copy()

        for gid in self.glacier_ids:
            gpos = (s.glaciers.glacier == gid) & (old_ice_thickness > 0.0)
            num_glacier_pixels = gpos.sum()

            if not gpos.any():
                continue

            logger.debug(f"Updating geometry for glacier {gid}")

            apply_default_delta_h = True

            # TODO: implement glacier advance

            if not apply_default_delta_h:
                continue

            update_pos_local = old_ice_thickness[gpos] >= MIN_ICE_THICKNESS
            update_pos = model.global_mask(update_pos_local, gpos)
            if update_pos_local.sum() < 2:
                # At least 2 pixels are required for calculating the normalized elevation range
                logger.debug("Less than 2 pixels available, skipping glacier")
                continue

            glacier_elevs = model.state.base.dem[gpos]
            glacier_mbs = mb[gpos]
            total_mb = glacier_mbs[update_pos_local].sum()
            glacier_area = num_glacier_pixels * model.grid.resolution**2  # (m2)
            terminus_elev = glacier_elevs.min()

            glacier_elevs_sorted_idxs = np.argsort(glacier_elevs)
            num_terminus_pixels = int(
                np.ceil(num_glacier_pixels * (TERMINUS_MASS_BALANCE_ELEVATION_PERCENTAGE / 100.0))
            )

            # Calculate the highest elevation of the lowest
            # TERMINUS_MASS_BALANCE_ELEVATION_PERCENTAGE percent of glacier pixels
            max_terminus_elev = glacier_elevs[glacier_elevs_sorted_idxs[:num_terminus_pixels]].max()

            # If the highest elevation of the lowest TERMINUS_ELEVATION_BUFFER m (by elevation) of
            # the glacier is higher, take this one instead
            max_terminus_elev = max(max_terminus_elev, terminus_elev + TERMINUS_ELEVATION_BUFFER)

            # Take the lowest mass balance value within this elevation range
            pos = model.global_mask(model.state.base.dem[gpos] <= max_terminus_elev, gpos)
            terminus_mb = mb[pos].min()

            max_allowed_change = min(terminus_mb, 0.0)  # MB must be negative
            mass_change = _ice_mass_change(
                glacier_elevs[update_pos_local],
                total_mb,
                glacier_area,
                max_allowed_change,
            )

            new_ice_thickness[update_pos] = (
                old_ice_thickness[update_pos] + (mass_change - firn_mb[update_pos]) / c.ICE_DENSITY
            ).clip(min=0.0)

        s.snow.thickness[CryoLayerID.ICE, :] = new_ice_thickness
        s.snow.ice_content[CryoLayerID.ICE, :] = new_ice_thickness * c.ICE_DENSITY

        model.snow.update_layers()

        # TODO: when glacier advance is enabled, update the previous ice mass distributions here


def _ice_mass_change(elevs, total_mb, glacier_area, max_allowed_change):
    """
    Calculate the distributed ice mass change for a single glacier.

    Parameters
    ----------
    elevs : ndarray(float)
        Glacier elevations (m).

    total_mb : float
        Total glacier surface mass balance (kg m-2).

    glacier_area : float
        Glacier area (m2).

    max_allowed_change : float
        Maximum allowed surface lowering (kg m-2).

    Returns
    -------
    mass_change : ndarray(float)
        Ice mass changes for the glacier pixels (kg m-2).

    References
    ----------
    .. [1] Huss, M., Jouvet, G., Farinotti, D., & Bauder, A. (2010). Future
       high-mountain hydrology: A new parameterization of glacier retreat.
       Hydrology and Earth System Sciences, 14(5), 815–829.
       https://doi.org/10.5194/hess-14-815-2010
    """
    if total_mb < 0:
        min_elev = elevs.min()
        max_elev = elevs.max()
        normalized_elevs = (max_elev - elevs) / (max_elev - min_elev)
        delta_h = _delta_h(normalized_elevs, glacier_area)
        fs = total_mb / delta_h.sum()
        mass_change = fs * delta_h

        # Restrict the surface lowering to max_allowed_change
        if max_allowed_change < 0:
            clippies = mass_change < max_allowed_change
            if clippies.any() and (~clippies).sum() > 1:
                total_we_change = mass_change.sum()
                mass_change[clippies] = max_allowed_change
                we_change_diff = total_we_change - mass_change.sum()
                mass_change[~clippies] += _mass_loss_surplus_distribution(
                    we_change_diff,
                    elevs[~clippies],
                )
    else:
        # If the mass balance is positive, distribute the mass gain equally over all glacier pixels
        mass_change = total_mb / elevs.size

    return mass_change


def _delta_h(h, glacier_area):
    if glacier_area >= 20e6:
        delta_h = -((h - 0.02) ** 6 + 0.12 * (h - 0.02))
    elif glacier_area >= 5e6:
        delta_h = -((h - 0.05) ** 4 + 0.19 * (h - 0.05) + 0.01)
    else:
        delta_h = -((h - 0.30) ** 2 + 0.60 * (h - 0.30) + 0.09)

    return delta_h.clip(-1, 0)


def _mass_loss_surplus_distribution(mass_loss, elevs):
    min_elev = elevs.min()
    max_elev = elevs.max()
    min_mass = 0.0

    x0 = 0.0
    x1 = mass_loss / elevs.size * 1.5
    y0 = np.sum(_scale_mass_loss_surplus(mass_loss, elevs, min_elev, max_elev, min_mass, x0))
    y1 = np.sum(_scale_mass_loss_surplus(mass_loss, elevs, min_elev, max_elev, min_mass, x1))

    k = (y1 - y0) / (x1 - x0)
    d = y0 - k * x0

    max_mass = (mass_loss - d) / k
    mass_loss_dist = _scale_mass_loss_surplus(
        mass_loss,
        elevs,
        min_elev,
        max_elev,
        min_mass,
        max_mass,
    )

    return mass_loss_dist


def _scale_mass_loss_surplus(mass_loss, elevs, min_elev, max_elev, min_mass, max_mass):
    return (max_mass - min_mass) * (elevs - min_elev) / (max_elev - min_elev) + min_mass



import subprocess

from openamundsen.util import to_yaml

from .conftest import base_config


def test_cli(tmp_path):
    config = base_config()
    config.start_date = "2020-01-18 00:00"
    config.end_date = "2020-01-18 00:00"
    config.results_dir = tmp_path

    config_file = tmp_path / "config.yml"
    with open(config_file, "w") as f:
        f.write(to_yaml(config.toDict()))

    subprocess.check_call(["openamundsen", str(config_file)])  # noqa: S603, S607



from numba import njit, prange

from openamundsen import constants as c
from openamundsen import heatconduction


def soil_properties(model):
    """
    Wrapper function for _soil_properties().
    """
    s = model.state

    _soil_properties(
        model.grid.roi_idxs,
        s.soil.thickness,
        s.soil.temp,
        s.soil.heat_cap,
        s.soil.vol_heat_cap_dry,
        s.soil.therm_cond,
        s.soil.therm_cond_dry,
        s.soil.vol_moisture_content,
        s.soil.vol_moisture_content_sat,
        s.soil.frac_frozen_moisture_content,
        s.soil.frac_unfrozen_moisture_content,
        s.soil.sat_water_pressure,
        s.soil.clapp_hornberger,
    )


@njit(parallel=True, cache=True)
def _soil_properties(
    roi_idxs,
    thickness,
    temp,
    heat_cap,
    vol_heat_cap_dry,
    therm_cond,
    therm_cond_dry,
    vol_moisture_content,
    vol_moisture_content_sat,
    frac_frozen_moisture_content,
    frac_unfrozen_moisture_content,
    sat_water_pressure,
    clapp_hornberger,
):
    """
    Calculate soil properties following [1] (equation numbers in the comments
    refer to this article).

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    thickness : ndarray(float, ndim=3)
        Soil thickness (m).

    temp : ndarray(float, ndim=3)
        Soil temperature (K).

    heat_cap : ndarray(float, ndim=3)
        Soil areal heat capacity (J K-1 m-2).

    vol_heat_cap_dry : ndarray(float, ndim=2)
        Volumetric heat capacity of dry soil (J K-1 m-3).

    therm_cond : ndarray(float, ndim=3)
        Soil thermal conductivity (W m-1 K-1).

    therm_cond_dry : ndarray(float, ndim=2)
        Dry soil thermal conductivity (W m-1 K-1).

    vol_moisture_content : ndarray(float, ndim=3)
        Volumetric soil moisture content (m3 m-3).

    vol_moisture_content_sat : ndarray(float, ndim=2)
        Volumetric soil moisture content at saturation (m3 m-3).

    frac_frozen_moisture_content : ndarray(float, ndim=3)
        Mass of frozen water as a fraction of that of liquid water at
        saturation.

    frac_unfrozen_moisture_content : ndarray(float, ndim=3)
        Mass of unfrozen water as a fraction of that of liquid water at
        saturation.

    sat_water_pressure : ndarray(float, ndim=2)
        Saturated soil water pressure (m).

    clapp_hornberger : ndarray(float, ndim=2)
        Clapp-Hornberger exponent.

    References
    ----------
    .. [1] Cox, P. M., Betts, R. A., Bunton, C. B., Essery, R. L. H., Rowntree,
       P. R., & Smith, J. (1999). The impact of new land surface physics on the
       GCM simulation of climate and climate sensitivity. Climate Dynamics, 15(3),
       183–203. https://doi.org/10.1007/s003820050276
    """
    num_pixels = len(roi_idxs)
    num_soil_layers = thickness.shape[0]

    # Calculate kappa (eq. (40))
    k = 1.0  # dimensionless constant depending on the nature of the soil
    kappa = k * (
        (c.ICE_DENSITY / c.WATER_DENSITY)
        * c.LATENT_HEAT_OF_FUSION
        / (c.T0 * c.GRAVITATIONAL_ACCELERATION)
    )

    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]

        for k in range(num_soil_layers):
            if vol_moisture_content[k, i, j] > 0:
                # Temperature above which all soil moisture is unfrozen (eq. (43))
                temp_max_frozen = (
                    c.T0
                    - sat_water_pressure[i, j]
                    / kappa
                    * (vol_moisture_content_sat[i, j] / vol_moisture_content[k, i, j])
                    ** clapp_hornberger[i, j]
                )

                if temp[k, i, j] < temp_max_frozen:
                    # Maximum unfrozen water that can exist at current soil temperature (eq. (39))
                    max_vol_unfrozen_moisture_content = vol_moisture_content_sat[i, j] * (
                        -kappa * (temp[k, i, j] - c.T0) / sat_water_pressure[i, j]
                    ) ** (-1 / clapp_hornberger[i, j])

                    # Actual amount of unfrozen water (limited by the total water content of the
                    # soil) (eqs. (41), (44))
                    vol_unfrozen_moisture_content = min(
                        max_vol_unfrozen_moisture_content,
                        vol_moisture_content[k, i, j],
                    )

                    # Change of unfrozen water concentration by temperature (eq. (45))
                    dTheta_u_by_dT = (
                        kappa
                        * vol_moisture_content_sat[i, j]
                        / (clapp_hornberger[i, j] * sat_water_pressure[i, j])
                        * (-kappa * (temp[k, i, j] - c.T0) / sat_water_pressure[i, j])
                        ** (-1 / clapp_hornberger[i, j] - 1)
                    )
                else:
                    vol_unfrozen_moisture_content = vol_moisture_content[k, i, j]
                    dTheta_u_by_dT = 0.0

                vol_frozen_moisture_content = (
                    (vol_moisture_content[k, i, j] - vol_unfrozen_moisture_content)
                    * c.WATER_DENSITY
                    / c.ICE_DENSITY
                )

                # Frozen/unfrozen moisture content (eqs. (48)-(49))
                frac_unfrozen_moisture_content[k, i, j] = (
                    vol_unfrozen_moisture_content / vol_moisture_content_sat[i, j]
                )
                frac_frozen_moisture_content[k, i, j] = (
                    (c.ICE_DENSITY / c.WATER_DENSITY)
                    * vol_frozen_moisture_content
                    / vol_moisture_content_sat[i, j]
                )

                # Volumetric heat capacity (eq. (37))
                vol_heat_cap = (
                    vol_heat_cap_dry[i, j]
                    + c.WATER_DENSITY * c.SPEC_HEAT_CAP_WATER * vol_unfrozen_moisture_content
                    + c.ICE_DENSITY * c.SPEC_HEAT_CAP_ICE * vol_frozen_moisture_content
                    + c.WATER_DENSITY
                    * (
                        (c.SPEC_HEAT_CAP_WATER - c.SPEC_HEAT_CAP_ICE) * temp[k, i, j]
                        + c.LATENT_HEAT_OF_FUSION
                    )
                    * dTheta_u_by_dT
                )

                # Areal heat capacity
                heat_cap[k, i, j] = vol_heat_cap * thickness[k, i, j]

                # Saturation concentration of liquid water for the current liquid water to
                # ice mass ratio (eq. (74))
                liquid_water_sat_conc = vol_moisture_content_sat[i, j] * (
                    frac_unfrozen_moisture_content[k, i, j]
                    / (
                        frac_unfrozen_moisture_content[k, i, j]
                        + frac_frozen_moisture_content[k, i, j]
                    )
                )

                # Thermal conductivity of saturated soil (eq. (75))
                therm_cond_sat = (
                    therm_cond_dry[i, j]
                    * c.THERM_COND_WATER**liquid_water_sat_conc
                    * c.THERM_COND_ICE ** (vol_moisture_content_sat[i, j] - liquid_water_sat_conc)
                    / c.THERM_COND_AIR ** vol_moisture_content_sat[i, j]
                )

                # Effective thermal conductivity (eq. (71))
                therm_cond[k, i, j] = (
                    therm_cond_sat - therm_cond_dry[i, j]
                ) * vol_moisture_content[k, i, j] / vol_moisture_content_sat[i, j] + therm_cond_dry[
                    i, j
                ]
            else:  # vol_moisture_content[k, i, j] == 0
                heat_cap[k, i, j] = vol_heat_cap_dry[i, j] * thickness[k, i, j]
                therm_cond[k, i, j] = therm_cond_dry[i, j]


def soil_temperature(model):
    """
    Update soil layer temperatures.

    Parameters
    ----------
    model : OpenAmundsen
        openAMUNDSEN model instance.

    References
    ----------
    .. [1] Cox, P. M., Betts, R. A., Bunton, C. B., Essery, R. L. H., Rowntree,
       P. R., & Smith, J. (1999). The impact of new land surface physics on the
       GCM simulation of climate and climate sensitivity. Climate Dynamics, 15(3),
       183–203. https://doi.org/10.1007/s003820050276
    """
    roi = model.grid.roi
    thickness = model.state.soil.thickness[:, roi]
    temp = model.state.soil.temp[:, roi]
    therm_cond = model.state.soil.therm_cond[:, roi]

    dT = heatconduction.temp_change_array(
        thickness,
        model.timestep,
        temp,
        therm_cond,
        temp[-1, :],
        thickness[-1, :],
        therm_cond[-1, :],
        model.state.soil.heat_flux[roi],
        model.state.soil.heat_cap[:, roi],
    )
    model.state.soil.temp[:, roi] += dT


def soil_heat_flux(model):
    """
    Wrapper function for _soil_heat_flux().
    """
    s = model.state

    _soil_heat_flux(
        model.grid.roi_idxs,
        s.surface.heat_flux,
        s.soil.heat_flux,
        s.soil.thickness,
        s.soil.temp,
        s.soil.therm_cond,
        s.snow.num_layers,
        s.snow.thickness,
        s.snow.temp,
        s.snow.therm_cond,
    )


@njit(parallel=True, cache=True)
def _soil_heat_flux(
    roi_idxs,
    surf_heat_flux,
    soil_heat_flux,
    soil_thickness,
    soil_temp,
    therm_cond_soil,
    num_snow_layers,
    snow_thickness,
    snow_temp,
    therm_cond_snow,
):
    """
    Calculate the soil heat flux, i.e., the flux from the surface to the top
    soil layer, following [1].

    Parameters
    ----------
    roi_idxs : ndarray(int, ndim=2)
        (N, 2)-array specifying the (row, col) indices within the data arrays
        that should be considered.

    surf_heat_flux : ndarray(float, ndim=2)
        Surface heat flux (W m-2).

    soil_heat_flux : ndarray(float, ndim=2)
        Soil heat flux (W m-2).

    soil_thickness : ndarray(float, ndim=3)
        Soil thickness (m).

    soil_temp : ndarray(float, ndim=3)
        Soil temperature (K).

    therm_cond_soil : ndarray(float, ndim=3)
        Soil thermal conductivity (W m-1 K-1).

    num_snow_layers : ndarray(float, ndim=2)
        Number of snow layers.

    snow_thickness : ndarray(float, ndim=3)
        Snow thickness (m).

    snow_temp : ndarray(float, ndim=3)
        Snow temperature (K).

    therm_cond_snow : ndarray(float, ndim=3)
        Snow thermal conductivity (W m-1 K-1).

    References
    ----------
    .. [1] Cox, P. M., Betts, R. A., Bunton, C. B., Essery, R. L. H., Rowntree,
       P. R., & Smith, J. (1999). The impact of new land surface physics on the
       GCM simulation of climate and climate sensitivity. Climate Dynamics, 15(3),
       183–203. https://doi.org/10.1007/s003820050276
    """
    num_pixels = len(roi_idxs)
    for idx_num in prange(num_pixels):
        i, j = roi_idxs[idx_num]

        ns = num_snow_layers[i, j]

        if ns == 0:
            soil_heat_flux[i, j] = surf_heat_flux[i, j]
        else:
            surf_moisture_conductance_bottom = 2.0 / (
                snow_thickness[ns - 1, i, j] / therm_cond_snow[ns - 1, i, j]
                + soil_thickness[0, i, j] / therm_cond_soil[0, i, j]
            )
            soil_heat_flux[i, j] = surf_moisture_conductance_bottom * (
                snow_temp[ns - 1, i, j] - soil_temp[0, i, j]
            )
